﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;

namespace PaintApp
{
	public class BitmapRectangle
	{
		public BitmapRectangle(Bitmap bitmap, Rectangle rectangle)
		{
			Bitmap = bitmap;
			Rectangle = rectangle;

			OldBitmap = new Bitmap(bitmap);
			OldX = rectangle.X;
			OldY = rectangle.Y;
			OldWidth = rectangle.Width;
			OldHeight = rectangle.Height;
		}

		public void UpdatePosition()
		{
			OldX = X;
			OldY = Y;
			OldWidth = Width;
			OldHeight = Height;
		}

		public Bitmap OldBitmap
		{
			private set;
			get;
		} = null;

		public int OldX
		{
			private set;
			get;
		}

		public int OldY
		{
			private set;
			get;
		}

		public int OldWidth
		{
			private set;
			get;
		}

		public int OldHeight
		{
			private set;
			get;
		}

		public void Dispose()
		{
			Bitmap.Dispose();
			if(OldBitmap != null)
				OldBitmap.Dispose();
		}

		public Bitmap Bitmap
		{
			set;
			get;
		}

		public Rectangle Rectangle
		{
			private set;
			get;
		}

		public int X
		{
			set
			{
				Rectangle rect = Rectangle;
				rect.X = value;
				Rectangle = rect;
			}
			get
			{
				return Rectangle.X;
			}
		}

		public int Y
		{
			set
			{
				Rectangle rect = Rectangle;
				rect.Y = value;
				Rectangle = rect;
			}
			get
			{
				return Rectangle.Y;
			}
		}

		public Point Location
		{
			set
			{
				Rectangle rect = Rectangle;
				rect.Location = value;
				Rectangle = rect;
			}
			get
			{
				return Rectangle.Location;
			}
		}

		public Size Size
		{
			set
			{
				Rectangle rect = Rectangle;
				rect.Size = value;
				Rectangle = rect;

				SetNewBitmapWhenSizeChange(new Size(rect.Width, rect.Height));
			}
			get
			{
				return Rectangle.Size;
			}
		}

		void SetNewBitmapWhenSizeChange(Size size)
		{
			Bitmap newBitmap = new Bitmap(size.Width, size.Height);
			Graphics g = Graphics.FromImage(newBitmap);
			g.DrawImage(OldBitmap, new Rectangle(0, 0, size.Width, size.Height), new Rectangle(0, 0, OldBitmap.Width, OldBitmap.Height), GraphicsUnit.Pixel);
			g.Dispose();
			Bitmap.Dispose();
			Bitmap = newBitmap;
		}

		public int Width
		{
			set
			{
				Rectangle rect = Rectangle;
				rect.Width = value;
				Rectangle = rect;

				SetNewBitmapWhenSizeChange(new Size(rect.Width, rect.Height));
			}
			get
			{
				return Rectangle.Width;
			}
		}

		public int Height
		{
			set
			{
				Rectangle rect = Rectangle;
				rect.Height = value;
				Rectangle = rect;

				SetNewBitmapWhenSizeChange(new Size(rect.Width, rect.Height));
			}
			get
			{
				return Rectangle.Height;
			}
		}
	}
}

