﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace PaintApp
{
	public partial class ColorReplaceForm : Form
	{
		public List<Color> Colors = new List<Color>();
		public Color BeforeColor = Color.Empty;
		public Color AfterColor = Color.Empty;
		public bool IsBeforeColorTransparent = false;
		public bool IsAfterColorTransparent = false;
		public int differenceR = 0;
		public int differenceG = 0;
		public int differenceB = 0;

		public ColorReplaceForm()
		{
			InitializeComponent();

			this.Load += ColorReplaceForm_Load;

			buttonOK.DialogResult = DialogResult.OK;
			buttonCancel.DialogResult = DialogResult.Cancel;

			buttonOK.Click += ButtonOK_Click;

			checkBox1.CheckedChanged += CheckBox1_CheckedChanged;
			checkBox2.CheckedChanged += CheckBox2_CheckedChanged;

			buttonSelectBeforeColor.Click += ButtonSelectBeforeColor_Click;
			buttonSelectAfterColor.Click += ButtonSelectAfterColor_Click;

			BeforeColorComboBox.SelectedIndexChanged += BeforeColorComboBox_SelectedIndexChanged;
			AfterColorComboBox.SelectedIndexChanged += AfterColorComboBox_SelectedIndexChanged;
		}

		private void ColorReplaceForm_Load(object sender, EventArgs e)
		{
			foreach(Color color in Colors)
			{
				string str = String.Format("R={0}, G={1}, B={2}, A={3}", color.R, color.G, color.B, color.A);
				BeforeColorComboBox.Items.Add(str);
				AfterColorComboBox.Items.Add(str);
			}
		}

		private void ButtonSelectBeforeColor_Click(object sender, EventArgs e)
		{
			ColorDialog dlg = new ColorDialog();
			if(dlg.ShowDialog() == DialogResult.OK)
			{
				pictureBox1.BackColor = dlg.Color;
				string str = String.Format("R={0}, G={1}, B={2}, A={3}", dlg.Color.R, dlg.Color.G, dlg.Color.B, dlg.Color.A);
				labelBeforeColor.Text = str;
			}
			dlg.Dispose();
		}

		private void ButtonSelectAfterColor_Click(object sender, EventArgs e)
		{
			ColorDialog dlg = new ColorDialog();
			if(dlg.ShowDialog() == DialogResult.OK)
			{
				pictureBox2.BackColor = dlg.Color;
				string str = String.Format("R={0}, G={1}, B={2}, A={3}", dlg.Color.R, dlg.Color.G, dlg.Color.B, dlg.Color.A);
				labelAfterColor.Text = str;
			}
			dlg.Dispose();
		}

		private void CheckBox1_CheckedChanged(object sender, EventArgs e)
		{
			if(checkBox1.Checked)
				pictureBox1.BackColor = Color.Empty;
		}

		private void CheckBox2_CheckedChanged(object sender, EventArgs e)
		{
			if(checkBox2.Checked)
				pictureBox2.BackColor = Color.Empty;
		}

		private void ButtonOK_Click(object sender, EventArgs e)
		{
			BeforeColor = pictureBox1.BackColor;
			AfterColor = pictureBox2.BackColor;

			if(checkBox1.Checked)
				IsBeforeColorTransparent = true;
			if(checkBox2.Checked)
				IsAfterColorTransparent = true;

			differenceR = (int)numericUpDownR.Value;
			differenceG = (int)numericUpDownG.Value;
			differenceB = (int)numericUpDownB.Value;
		}

		private void BeforeColorComboBox_SelectedIndexChanged(object sender, EventArgs e)
		{
			int index = BeforeColorComboBox.SelectedIndex;
			pictureBox1.BackColor = Colors[index];
		}

		private void AfterColorComboBox_SelectedIndexChanged(object sender, EventArgs e)
		{
			int index = AfterColorComboBox.SelectedIndex;
			pictureBox2.BackColor = Colors[index];
		}

		private void buttonSelectBeforeColor_Click_1(object sender, EventArgs e)
		{

		}
	}
}
