﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.Drawing.Imaging;

namespace PaintApp
{
	public partial class Form1 : Form
	{
		public Form1()
		{
			InitializeComponent();

			UndoRedoStatusLabel.Text = "";
			BitmapSizeStatusLabel.Text = "";
			MousePositionStatusLabel.Text = "";
			MousePositionColorStatusLabel.Text = "";

			userControlImage1.BitmapChanged += UserControlImage1_BitmapChanged;

			userControlImage1.PictureBoxMouseDown += UserControlImage1_PictureBoxMouseDown;

		    userControlColorBox1.Anchor = AnchorStyles.Top | AnchorStyles.Right;
		    userControlColorBox2.Anchor = AnchorStyles.Top | AnchorStyles.Right;
		    userControlColorBox3.Anchor = AnchorStyles.Top | AnchorStyles.Right;

			pictureBoxSetColor.BackColor = userControlImage1.DrawColor;

			ShowInitBitmap();

			userControlImage1.AllowDrop = true;
            userControlImage1.DragOver += UserControlImage1_DragOver;
            userControlImage1.DragDrop += UserControlImage1_DragDrop;

			Text = Application.ProductName + " Ver." + Application.ProductVersion;

            userControlImage1.PictureBoxMouseMove += UserControlImage1_PictureBoxMouseMove;
		}

		void ShowPointInfo(int x, int y)
		{
			Bitmap bitmap = userControlImage1.Bitmap;

			if(bitmap != null && userControlImage1.BitmapRectangle != null)
			{
				// 範囲選択されているときは選択範囲の座標、幅、高さを表示する
				BitmapRectangle rect = userControlImage1.BitmapRectangle;
				MousePositionStatusLabel.Text = String.Format("( {0}, {1}) ( {2}, {3})", rect.X, rect.Y, rect.Width, rect.Height);

				// BackColorを反映させるためにはなにかをTextに設定する必要がある
				MousePositionColorStatusLabel.Text = "    ";

				// 範囲選択されているときは左上の座標の色を表示する
				if(0 <= rect.X && rect.X < userControlImage1.Bitmap.Width && 0 <= rect.Y && rect.Y < userControlImage1.Bitmap.Height)
					MousePositionColorStatusLabel.BackColor = userControlImage1.GetColorFromPoint(rect.X, rect.Y);
				else
					MousePositionColorStatusLabel.BackColor = Color.Empty;
			}
			else if(bitmap != null && x < userControlImage1.Bitmap.Width && y < userControlImage1.Bitmap.Height)
			{
				// 範囲選択されていないときはマウスポインタの位置を表示する
				MousePositionStatusLabel.Text = String.Format("( {0}, {1})", x, y);
				MousePositionColorStatusLabel.Text = "    ";
				MousePositionColorStatusLabel.BackColor = userControlImage1.GetColorFromPoint(x, y);
			}
			else
			{
				// マウスポインタの位置がビットマップの外であるなら「-」を表示する
				MousePositionStatusLabel.Text = "-";
				MousePositionColorStatusLabel.Text = "    ";
				MousePositionColorStatusLabel.BackColor = Color.Empty;
			}
		}

		private void UserControlImage1_PictureBoxMouseMove(object sender, UserControlImage.PictureBoxMouseMoveEventArgs e)
        {
			ShowPointInfo(e.X, e.Y);
		}

		private void UserControlImage1_DragDrop(object sender, DragEventArgs e)
        {
			if(e.Data.GetDataPresent(DataFormats.FileDrop))
			{
				string[] filepath = (string[])e.Data.GetData(DataFormats.FileDrop);
				if(filepath.Length == 1)
				{
					if(userControlImage1.LoadImageFile(filepath[0]))
					{
						filePath = filepath[0];
						userControlImage1.EditMode = EditMode.Selection;
					}
					else
						MessageBox.Show("これは画像ファイルではありません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				}
			}
		}

        private void UserControlImage1_DragOver(object sender, DragEventArgs e)
        {
			if(e.Data.GetDataPresent(DataFormats.FileDrop))
				e.Effect = DragDropEffects.All;
			else
				e.Effect = DragDropEffects.None;
		}

        private void UserControlImage1_BitmapChanged(object sender, EventArgs e)
		{
			ShowBitmapInfo();
		}

		void ShowBitmapInfo()
		{
			int undoCount = userControlImage1.UndoBufCount;
			int redoCount = userControlImage1.RedoBufCount;

			if(userControlImage1.Bitmap != null)
			{
				UndoRedoStatusLabel.Text = String.Format("Undo ={0}  Redo={1}", undoCount, redoCount);
				BitmapSizeStatusLabel.Text = String.Format("{0} × {1} （{2}倍）",
						userControlImage1.Bitmap.Width, userControlImage1.Bitmap.Height, userControlImage1.DisplayMagnification);
			}

			//MousePositionStatusLabel;
			//MousePositionColorStatusLabel;
		}

		void ShowInitBitmap()
		{
			Bitmap bmp = new Bitmap(200, 200);
			Graphics g = Graphics.FromImage(bmp);
			g.Clear(Color.White);
			g.Dispose();
			userControlImage1.Bitmap = bmp;
			userControlImage1.ShowBitmap(bmp);
		}

		private void UserControlImage1_PictureBoxMouseDown(object sender, UserControlImage.PictureBox1MouseDownArgs e)
		{
			if(userControlColorBox1.IsMemoryColor)
				userControlColorBox1.Color = e.Color;
			if(userControlColorBox2.IsMemoryColor)
				userControlColorBox2.Color = e.Color;
			if(userControlColorBox3.IsMemoryColor)
				userControlColorBox3.Color = e.Color;

			if(userControlImage1.EditMode == EditMode.FillColor)
				userControlImage1.FillColor(new Point(e.X, e.Y), userControlImage1.DrawColor);
		}

		private void Form1_Load(object sender, EventArgs e)
		{
		
		}

		// 現在編集中の画像ファイルのパス
		string filePath = "";

		void OpenFile()
		{
			OpenFileDialog dlg = new OpenFileDialog();
			dlg.CheckFileExists = true;
			dlg.CheckPathExists = true;
			dlg.Filter = "画像ファイル(*.bmp; *.jpg; *.png)|*.bmp; *.jpg; *.png|すべてのファイル(*.*)|*.*";
			if(dlg.ShowDialog() == DialogResult.OK)
			{
				if(userControlImage1.LoadImageFile(dlg.FileName))
				{
					filePath = dlg.FileName;
					CheckMenu1(SelectionMenuItem);
				}
			}
			dlg.Dispose();
		}

		private void OpenFileMenuItem_Click(object sender, EventArgs e)
		{
			OpenFile();
		}

		void SaveAsFile()
		{
			SaveFileDialog dlg = new SaveFileDialog();
			dlg.Filter = "画像ファイル(*.bmp)|*.bmp|画像ファイル(*.jpg)|*.jpg|画像ファイル(*.png)|*.png|すべてのファイル(*.*)|*.*";
			if(dlg.ShowDialog() == DialogResult.OK)
			{
				userControlImage1.Bitmap.Save(dlg.FileName);
				filePath = dlg.FileName;
			}
			dlg.Dispose();
		}

		private void SaveAsFileMenuItem_Click(object sender, EventArgs e)
		{
			SaveAsFile();
		}

		void SaveFile()
		{
			if(File.Exists(filePath))
			{
				userControlImage1.Bitmap.Save(filePath);
			}
			else
			{
				SaveAsFile();
			}
		}

		private void SaveFileMenuItem_Click(object sender, EventArgs e)
		{
			SaveFile();
		}

		private void SizeChangeMenuItem_Click(object sender, EventArgs e)
		{
			SizeChange();
		}

		void SizeChange()
		{
			Bitmap bmp = userControlImage1.Bitmap;
			if(bmp == null)
				return;

			SizeChangeForm form = new SizeChangeForm();
			form.bmpWidth = bmp.Width;
			form.bmpHeight = bmp.Height;
			form.Text = "サイズ変更";

			if(form.ShowDialog() == DialogResult.OK)
			{
				userControlImage1.SizeChange(form.newWidth, form.newHeight);
			}

			form.Dispose();
		}

		private void ReverseLeftRightMenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.RotateFlip(RotateFlipType.RotateNoneFlipX);
			if(!ret)
				MessageBox.Show("画像が読み込まれていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void ReverseUpperLowerMenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.RotateFlip(RotateFlipType.RotateNoneFlipY);
			if(!ret)
				MessageBox.Show("画像が読み込まれていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void Rotate90MenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.RotateFlip(RotateFlipType.Rotate90FlipNone);
			if(!ret)
				MessageBox.Show("画像が読み込まれていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void Rotate180MenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.RotateFlip(RotateFlipType.Rotate180FlipNone);
			if(!ret)
				MessageBox.Show("画像が読み込まれていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void Rotate270MenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.RotateFlip(RotateFlipType.Rotate270FlipNone);
			if(!ret)
				MessageBox.Show("画像が読み込まれていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void SelectionReverseLeftRightMenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.SelectionRotateFlip(RotateFlipType.RotateNoneFlipX);
			if(!ret)
				MessageBox.Show("画像が読み込まれていないか範囲選択がされていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void SelectionReverseUpperLowerMenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.SelectionRotateFlip(RotateFlipType.RotateNoneFlipY);
			if(!ret)
				MessageBox.Show("画像が読み込まれていないか範囲選択がされていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void SelectionRotate90MenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.SelectionRotateFlip(RotateFlipType.Rotate90FlipNone);
			if(!ret)
				MessageBox.Show("画像が読み込まれていないか範囲選択がされていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void SelectionRotate180MenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.SelectionRotateFlip(RotateFlipType.Rotate180FlipNone);
			if(!ret)
				MessageBox.Show("画像が読み込まれていないか範囲選択がされていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void SelectionRotate270MenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.SelectionRotateFlip(RotateFlipType.Rotate270FlipNone);
			if(!ret)
				MessageBox.Show("画像が読み込まれていないか範囲選択がされていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void ColorReplaceMenuItem_Click(object sender, EventArgs e)
		{
			if(userControlImage1.Bitmap == null)
			{
				MessageBox.Show("画像が読み込まれていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}

			ColorReplaceForm form = new ColorReplaceForm();
			form.Text = "色の置換（画像全体）";

			List<Color> colors = new List<Color>();
			if(userControlColorBox1.Color != Color.Empty)
				colors.Add(userControlColorBox1.Color);
			if(userControlColorBox2.Color != Color.Empty)
				colors.Add(userControlColorBox2.Color);
			if(userControlColorBox3.Color != Color.Empty)
				colors.Add(userControlColorBox3.Color);
			form.Colors = colors;

			if(form.ShowDialog() == DialogResult.OK)
			{
				userControlImage1.ReplaceColor(form.BeforeColor, form.AfterColor,
						form.IsBeforeColorTransparent, form.IsAfterColorTransparent, form.differenceR, form.differenceG, form.differenceB);
			}
			form.Dispose();
		}

		Bitmap GetSelectionBitmap()
		{
			Bitmap bmp = userControlImage1.Bitmap;
			Rectangle rectSrc = userControlImage1.DragDropRectangle;

			Bitmap trimingBitmap = new Bitmap(rectSrc.Width, rectSrc.Height);
			Graphics g = Graphics.FromImage(trimingBitmap);
			g.DrawImage(bmp, 0, 0, rectSrc, GraphicsUnit.Pixel);
			g.Dispose();

			return trimingBitmap;
		}

		Bitmap SetSelectionBitmap(Bitmap trimingBitmap, Rectangle rect)
		{
			Bitmap bmp = userControlImage1.Bitmap;

			Bitmap newBmp = new Bitmap(bmp.Width, bmp.Height);
			Graphics g = Graphics.FromImage(newBmp);
			g.DrawImage(bmp, 0, 0, bmp.Width, bmp.Height);
			g.DrawImage(trimingBitmap, rect);
			g.Dispose();
			
			return newBmp;
		}

		Bitmap SetSelectionTransparent(Rectangle rect, Color beforeColor, int differenceR, int differenceG, int differenceB)
		{
			Bitmap bmp = userControlImage1.Bitmap;

			int minR = beforeColor.R - differenceR;
			int maxR = beforeColor.R + differenceR;
			int minG = beforeColor.G - differenceG;
			int maxG = beforeColor.G + differenceG;
			int minB = beforeColor.B - differenceB;
			int maxB = beforeColor.B + differenceB;

			if(minR < 0)
				minR = 0;
			if(maxR > 255)
				maxR = 255;
			if(minG < 0)
				minG = 0;
			if(maxG > 255)
				maxG = 255;
			if(minB < 0)
				minB = 0;
			if(maxB > 255)
				maxB = 255;

			int width = bmp.Width;
			int height = bmp.Height;

			Bitmap newBmp = new Bitmap(width, height);

			for(int x = 0; x < width; x++)
			{
				for(int y = 0; y < height; y++)
				{
					Color color = bmp.GetPixel(x, y);

					if((x < rect.Left || rect.Right < x) || (y < rect.Top || rect.Bottom < y))
					{
						newBmp.SetPixel(x, y, color);
					}
					else
					{
						if(minR <= color.R && color.R <= maxR && minG <= color.G && color.G <= maxG && minB <= color.B && color.B <= maxB)
							newBmp.SetPixel(x, y, Color.FromArgb(0, 0, 0, 0));
						else
							newBmp.SetPixel(x, y, color);
					}
				}
			}
			return newBmp;
		}

		Bitmap GetBitmapReplaceColor(Bitmap sourceBmp, Color beforeColor, Color afterColor, int differenceR, int differenceG, int differenceB)
		{
			int width = sourceBmp.Width;
			int height = sourceBmp.Height;
			Bitmap newBitmap = new Bitmap(width, height);

			int minR = beforeColor.R - differenceR;
			int maxR = beforeColor.R + differenceR;
			int minG = beforeColor.G - differenceG;
			int maxG = beforeColor.G + differenceG;
			int minB = beforeColor.B - differenceB;
			int maxB = beforeColor.B + differenceB;

			if(minR < 0)
				minR = 0;
			if(maxR > 255)
				maxR = 255;
			if(minG < 0)
				minG = 0;
			if(maxG > 255)
				maxG = 255;
			if(minB < 0)
				minB = 0;
			if(maxB > 255)
				maxB = 255;

			for(int x = 0; x < width; x++)
			{
				for(int y = 0; y < height; y++)
				{
					Color color = sourceBmp.GetPixel(x, y);

					// 透明と黒を混同しないようにする
					if(beforeColor != Color.FromArgb(0, 0, 0, 0) && color != Color.FromArgb(0, 0, 0, 0) && minR <= color.R && color.R <= maxR && minG <= color.G && color.G <= maxG && minB <= color.B && color.B <= maxB)
						newBitmap.SetPixel(x, y, afterColor);
					else if(beforeColor == Color.FromArgb(0, 0, 0, 0) && color == Color.FromArgb(0, 0, 0, 0))
						newBitmap.SetPixel(x, y, afterColor);
					else
						newBitmap.SetPixel(x, y, color);
				}
			}
			return newBitmap;
		}

		private void SelectionColorReplaceMenuItem_Click(object sender, EventArgs e)
		{
			if(userControlImage1.Bitmap == null)
			{
				MessageBox.Show("画像が読み込まれていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}

			if(!userControlImage1.IsSelected())
			{
				MessageBox.Show("範囲選択されていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}

			ColorReplaceForm form = new ColorReplaceForm();
			form.Text = "色の置換（選択されている部分のみ）";

			List<Color> colors = new List<Color>();
			if(userControlColorBox1.Color != Color.Empty)
				colors.Add(userControlColorBox1.Color);
			if(userControlColorBox2.Color != Color.Empty)
				colors.Add(userControlColorBox2.Color);
			if(userControlColorBox3.Color != Color.Empty)
				colors.Add(userControlColorBox3.Color);
			form.Colors = colors;

			if(form.ShowDialog() == DialogResult.OK)
			{
				userControlImage1.ReplaceColorInRectangle(
					form.BeforeColor, form.AfterColor, form.IsBeforeColorTransparent, form.IsAfterColorTransparent, form.differenceR, form.differenceG, form.differenceB);
			}
			form.Dispose();
		}

		private void ColorReverseMenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.ReverseColor();
			if(!ret)
				MessageBox.Show("画像が読み込まれていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void SelectionColorReverseMenuItem_Click(object sender, EventArgs e)
		{
			bool ret = userControlImage1.ReverseColorInRectangle();

			if(!ret)
				MessageBox.Show("画像が読み込まれていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		void CheckMenu1(ToolStripMenuItem menuItem)
		{
			SelectionMenuItem.Checked = false;
			SelectionFreeMenuItem.Checked = false;

			DrawFigureMenuItem.Checked = false;

			DrawRectangleMenuItem.Checked = false;
			FillRectangleMenuItem.Checked = false;
			DrawEllipseMenuItem.Checked = false;
			FillEllipseMenuItem.Checked = false;
			DrawLineMenuItem.Checked = false;
			FreeLineMenuItem.Checked = false;
			FillColorMenuItem.Checked = false;

			if(menuItem == SelectionMenuItem)
				SelectionMenuItem.Checked = true;
			else if(menuItem == SelectionFreeMenuItem)
				SelectionFreeMenuItem.Checked = true;
			else if(menuItem == DrawRectangleMenuItem)
			{
				DrawFigureMenuItem.Checked = true;
				DrawRectangleMenuItem.Checked = true;
			}
			else if(menuItem == FillRectangleMenuItem)
			{
				DrawFigureMenuItem.Checked = true;
				FillRectangleMenuItem.Checked = true;
			}
			else if(menuItem == DrawEllipseMenuItem)
			{
				DrawFigureMenuItem.Checked = true;
				DrawEllipseMenuItem.Checked = true;
			}
			else if(menuItem == FillEllipseMenuItem)
			{
				DrawFigureMenuItem.Checked = true;
				FillEllipseMenuItem.Checked = true;
			}
			else if(menuItem == DrawLineMenuItem)
			{
				DrawFigureMenuItem.Checked = true;
				DrawLineMenuItem.Checked = true;
			}
			else if(menuItem == FreeLineMenuItem)
			{
				DrawFigureMenuItem.Checked = true;
				FreeLineMenuItem.Checked = true;
			}
			else if(menuItem == FillColorMenuItem)
			{
				FillColorMenuItem.Checked = true;
			}
		}

		private void DrawLineMenuItem_Click(object sender, EventArgs e)
		{
			CheckMenu1(DrawLineMenuItem);
			userControlImage1.EditMode = EditMode.DrawLine;
		}

		private void SelectionFreeMenuItem_Click(object sender, EventArgs e)
		{
			CheckMenu1(SelectionFreeMenuItem);
			userControlImage1.EditMode = EditMode.SelectionFree;
		}

		private void FreeLineMenuItem_Click(object sender, EventArgs e)
		{
			CheckMenu1(FreeLineMenuItem);
			userControlImage1.EditMode = EditMode.FreeLine;
		}

		private void SelectionMenuItem_Click(object sender, EventArgs e)
		{
			CheckMenu1(SelectionMenuItem);
			userControlImage1.EditMode = EditMode.Selection;
		}

		private void DrawRectangleMenuItem_Click(object sender, EventArgs e)
		{
			CheckMenu1(DrawRectangleMenuItem);
			userControlImage1.EditMode = EditMode.DrawRectangle;
		}

		private void FillRectangleMenuItem_Click(object sender, EventArgs e)
		{
			CheckMenu1(FillRectangleMenuItem);
			userControlImage1.EditMode = EditMode.FillRectangle;
		}

		private void DrawEllipseMenuItem_Click(object sender, EventArgs e)
		{
			CheckMenu1(DrawEllipseMenuItem);
			userControlImage1.EditMode = EditMode.DrawEllipse;
		}

		private void FillEllipseMenuItem_Click(object sender, EventArgs e)
		{
			CheckMenu1(FillEllipseMenuItem);
			userControlImage1.EditMode = EditMode.FillEllipse;
		}

		private void buttonSetColor_Click(object sender, EventArgs e)
		{
			ColorDialog dlg = new ColorDialog();
			if(dlg.ShowDialog() == DialogResult.OK)
			{
				pictureBoxSetColor.BackColor = dlg.Color;
				userControlImage1.DrawColor = dlg.Color;
			}
			dlg.Dispose();
		}

		private void UndoMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.Undo();
		}

		private void RedoMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.Redo();
		}

		private void UndoRedoMenuItem_DropDownOpening(object sender, EventArgs e)
		{
			UndoMenuItem.Enabled = true;
			RedoMenuItem.Enabled = true;

			if(!userControlImage1.CanUndo())
				UndoMenuItem.Enabled = false;
			if(!userControlImage1.CanRedo())
				RedoMenuItem.Enabled = false;
		}

		private void InsertStringMenuItem_Click(object sender, EventArgs e)
		{
			InsertStringForm form = new InsertStringForm();
			if(form.ShowDialog() == DialogResult.OK && form.String != "")
			{
				Bitmap bitmap = userControlImage1.Bitmap;
				if(bitmap != null)
				{
					string str = form.String;
					Font font = form.StringFont;
					Point point = new Point(0, 0);
					Color color = form.Color;

					bool ret = userControlImage1.InsertString(str, font, color, point);
					if(!ret)
						MessageBox.Show("画像が読み込まれていないか文字列が設定されていません", "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
				}
			}
			form.Dispose();
		}

		private void デバッグ情報表示ToolStripMenuItem_Click(object sender, EventArgs e)
		{
		}

		private void CopyMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.CopySelection();
			
		}

		private void CutMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.CutSelection();
		}

		private void PasteMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.PasteImage();
		}

		private void FillColorMenuItem_Click(object sender, EventArgs e)
		{
			CheckMenu1(FillColorMenuItem);
			userControlImage1.EditMode = EditMode.FillColor;
		}


		private void SelectionMoveUpMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.SelectionMoveUp();
		}

		private void SelectionMoveDownMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.SelectionMoveDown();
		}

		private void SelectionMoveLeftMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.SelectionMoveLeft();
		}

		private void SelectionMoveRightMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.SelectionMoveRight();
		}

		private void SelectAllMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.SelectAll();
		}

		protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
		{
			if(keyData == Keys.Escape)
			{
				userControlImage1.ClearBitmapRectangle(true);
				return true;
			}
			if(keyData == Keys.Up)
			{
				userControlImage1.OnKeyDownUp();
				if(userControlImage1.BitmapRectangle != null)
					ShowPointInfo(userControlImage1.BitmapRectangle.X, userControlImage1.BitmapRectangle.Y);
				return true;
			}

			if(keyData == Keys.Down)
			{
				userControlImage1.OnKeyDownDown();
				if(userControlImage1.BitmapRectangle != null)
					ShowPointInfo(userControlImage1.BitmapRectangle.X, userControlImage1.BitmapRectangle.Y);
				return true;
			}

			if(keyData == Keys.Right)
			{
				userControlImage1.OnKeyDownRight();
				if(userControlImage1.BitmapRectangle != null)
					ShowPointInfo(userControlImage1.BitmapRectangle.X, userControlImage1.BitmapRectangle.Y);
				return true;
			}

			if(keyData == Keys.Left)
			{
				userControlImage1.OnKeyDownLeft();
				if(userControlImage1.BitmapRectangle != null)
					ShowPointInfo(userControlImage1.BitmapRectangle.X, userControlImage1.BitmapRectangle.Y);
				return true;
			}

			return base.ProcessCmdKey(ref msg, keyData);
		}

		private void PasteDefaultMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.PasteMode1 = PasteMode1.Default;
		}

		private void PasteUpperSideMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.PasteMode1 = PasteMode1.UpperSide;
		}

		private void PasteUnderSideMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.PasteMode1 = PasteMode1.UnderSide;
		}

		private void PasteLeftSideMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.PasteMode1 = PasteMode1.LeftSide;
		}

		private void PasteRightSideMenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.PasteMode1 = PasteMode1.RightSide;
		}

		void CheckMenuPasteMode()
		{
			PasteMode1 mode = userControlImage1.PasteMode1;

			PasteDefaultMenuItem.Checked = false;
			PasteLeftSideMenuItem.Checked = false;
			PasteRightSideMenuItem.Checked = false;
			PasteUpperSideMenuItem.Checked = false;
			PasteUnderSideMenuItem.Checked = false;

			if(mode == PasteMode1.Default)
				PasteDefaultMenuItem.Checked = true;
			if(mode == PasteMode1.LeftSide)
				PasteLeftSideMenuItem.Checked = true;
			if(mode == PasteMode1.RightSide)
				PasteRightSideMenuItem.Checked = true;
			if(mode == PasteMode1.UpperSide)
				PasteUpperSideMenuItem.Checked = true;
			if(mode == PasteMode1.UnderSide)
				PasteUnderSideMenuItem.Checked = true;
		}

		private void CopyPasteMenuItem_DropDownOpening(object sender, EventArgs e)
		{
			CheckMenuPasteMode();
			CheckMenuPasteAutoSize();
		}

		void CheckMenuPasteAutoSize()
		{
			if(userControlImage1.IsAutoSizePaste)
				IsAutoSizePasteMenuItem.Checked = true;
			else
				IsAutoSizePasteMenuItem.Checked = false;
		}

		private void IsAutoSizePasteMenuItem_Click(object sender, EventArgs e)
        {
			if(userControlImage1.IsAutoSizePaste == true)
				userControlImage1.IsAutoSizePaste = false;
			else
				userControlImage1.IsAutoSizePaste = true;
		}

        private void Show100MenuItem_Click(object sender, EventArgs e)
        {
			userControlImage1.DisplayMagnification = 1.0F;
			ShowBitmapInfo();
		}

        private void Show75MenuItem_Click(object sender, EventArgs e)
        {
			userControlImage1.DisplayMagnification = 0.75F;
			ShowBitmapInfo();
		}

		private void Show50MenuItem_Click(object sender, EventArgs e)
        {
			userControlImage1.DisplayMagnification = 0.5F;
			ShowBitmapInfo();
		}

		private void Show150MenuItem_Click(object sender, EventArgs e)
        {
			userControlImage1.DisplayMagnification = 1.5F;
			ShowBitmapInfo();
		}

		private void Show200MenuItem_Click(object sender, EventArgs e)
		{
			userControlImage1.DisplayMagnification = 2.0F;
			ShowBitmapInfo();
		}

		private void Show300MenuItem_Click(object sender, EventArgs e)
        {
			userControlImage1.DisplayMagnification = 3.0F;
			ShowBitmapInfo();
		}

		private void DisplayMagnificationMenuItem_DropDownOpening(object sender, EventArgs e)
        {
			Show100MenuItem.Checked = false;
			Show75MenuItem.Checked = false;
			Show50MenuItem.Checked = false;
			Show150MenuItem.Checked = false;
			Show200MenuItem.Checked = false;
			Show300MenuItem.Checked = false;

			float f = userControlImage1.DisplayMagnification;

			if(f == 1.0F)
				Show100MenuItem.Checked = true;
			if(f == 0.75F)
				Show75MenuItem.Checked = true;
			if(f == 0.5F)
				Show50MenuItem.Checked = true;
			if(f == 1.5F)
				Show150MenuItem.Checked = true;
			if(f == 2.0F)
				Show200MenuItem.Checked = true;
			if(f == 3.0F)
				Show300MenuItem.Checked = true;
		}

        private void aaaToolStripMenuItem_Click(object sender, EventArgs e)
        {
			ShowInitBitmap();
        }

        private void UndoRedoMenuItem_Click(object sender, EventArgs e)
        {

        }

        private void 編集ToolStripMenuItem_DropDownOpening(object sender, EventArgs e)
        {
			UndoMenuItem.Visible = true;
			RedoMenuItem.Visible = true;

			if(!userControlImage1.CanUndo())
				UndoMenuItem.Visible = false;
			if(!userControlImage1.CanRedo())
				RedoMenuItem.Visible = false;

			CheckMenuPasteMode();
			CheckMenuPasteAutoSize();
		}

        private void SelectRectangleFromDialogMenuItem_Click(object sender, EventArgs e)
        {
			SelectRectangleForm form = new SelectRectangleForm();
			form.userControlImage = this.userControlImage1;
			form.ShowDialog();
		}
    }
}
