﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Reflection;

namespace PaintApp
{
	public partial class UserControlImage : UserControl
	{
		public UserControlImage()
		{
			InitializeComponent();

			// スクロールバーがスクロールされたときのイベントハンドラ
			this.vScrollBar1.Scroll += VScrollBar1_Scroll;
			this.hScrollBar1.Scroll += HScrollBar1_Scroll;

			this.pictureBox1.MouseDown += PictureBox1_MouseDown;
			this.pictureBox1.MouseUp += PictureBox1_MouseUp;
			this.pictureBox1.MouseMove += PictureBox1_MouseMove;

			this.Resize += UserControlImage_Resize;

			this.pictureBox1.Paint += PictureBox1_Paint;
		}

		public void SelectRectangle(int x, int y, int width, int height)
		{
			if(BitmapRectangle != null)
				UniteBitmapRectangle(new Point(0, 0));

			startPoint = new Point(x, y);
			endPoint = new Point(x + width, y + height);

			SelectRectangle();

			InvalidatePictureBox();
		}

		float displayMagnification = 1.0F;

		public float DisplayMagnification
		{
			get { return displayMagnification; }
			set
			{
				displayMagnification = value;
				pictureBox1.Invalidate();
			}
		}

		public bool IsAutoSizePaste
		{
			set;
			get;
		}
		= false;

		public void SelectAll()
		{
			this.UniteBitmapRectangle(new Point(0, 0));

			BitmapRectangle = new BitmapRectangle(new Bitmap(Bitmap), new Rectangle(0, 0, Bitmap.Width, Bitmap.Height));
			CreateBlankBitmap(new Rectangle(0, 0, Bitmap.Width, Bitmap.Height));

			Bitmap bitmap = DrawBoderRectangle(Bitmap);
			ShowBitmap(bitmap);
			bitmap.Dispose();
		}

		public bool SelectionMove(int moveX, int moveY)
		{
			if(BitmapRectangle == null)
				return false;

			if(moveX != 0)
				BitmapRectangle.X += moveX;
			if(moveY != 0)
				BitmapRectangle.Y += moveY;
			
			Bitmap bitmap = new Bitmap(Bitmap);
			Graphics g = Graphics.FromImage(bitmap);

			if(blankBitmap != null)
				g.DrawImage(blankBitmap.Bitmap, blankBitmap.Location);

			if(blankColorPoints != null)
			{
				foreach(ColorPoint cp in blankColorPoints)
					bitmap.SetPixel(cp.X, cp.Y, cp.Color);
			}

			g.DrawImage(BitmapRectangle.Bitmap, BitmapRectangle.Location);
			g.Dispose();

			Bitmap bitmap1 = DrawBoderRectangle(bitmap);
			bitmap.Dispose();
			ShowBitmap(bitmap1);
			bitmap1.Dispose();
			return true;
		}

		public bool OnKeyDownUp()
		{
			if(!SelectionMoveUp())
				VScrollMove(-1);
			return true;
		}

		public bool OnKeyDownDown()
		{
			if(!SelectionMoveDown())
				VScrollMove(1);
			return true;
		}

		public bool OnKeyDownLeft()
		{
			if(!SelectionMoveLeft())
				HScrollMove(-1);
			return true;
		}

		public bool OnKeyDownRight()
		{
			if(!SelectionMoveRight())
				HScrollMove(1);
			return true;
		}

		public bool SelectionMoveUp()
		{
			return SelectionMove(0, -1);
		}

		public bool SelectionMoveDown()
		{
			return SelectionMove(0, 1);
		}

		public bool SelectionMoveLeft()
		{
			return SelectionMove(-1, 0);
		}

		public bool SelectionMoveRight()
		{
			return SelectionMove(1, 0);
		}

		public bool VScrollMove(int i)
		{
			int value = vScrollBar1.Value;
			if(vScrollBar1.Minimum <= value + i && value + i <= vScrollBar1.Maximum)
			{
				this.vScrollBar1.Value = value + i;
				ShowBitmap(Bitmap);
				return true;
			}
			return false;
		}

		public bool HScrollMove(int i)
		{
			int value = hScrollBar1.Value;
			if(hScrollBar1.Minimum <= value + i && value + i <= hScrollBar1.Maximum)
			{
				this.hScrollBar1.Value = value + i;
				ShowBitmap(Bitmap);
				return true;
			}
			return false;
		}

		public bool InsertString(string str, Font font, Color color, Point point)
		{
			if(Bitmap == null)
				return false;

			if(str == "")
				return false;

			// 文字列が描画される矩形を求める
			Rectangle rect = GetStringRectangle(str, font, point);

			Bitmap stringBitmap = new Bitmap(rect.Width, rect.Height);
			Graphics g = Graphics.FromImage(stringBitmap);
			g.DrawString(str, font, new SolidBrush(color), point);
			g.Dispose();
			BitmapRectangle = new BitmapRectangle(stringBitmap, rect);

			EditMode = EditMode.Selection;
			InvalidatePictureBox();
			return true;
		}

		Rectangle GetStringRectangle(string str, Font font, Point point)
		{
			Bitmap bitmap = Bitmap;
			Graphics g = Graphics.FromImage(bitmap);
			StringFormat sf = new StringFormat();
			SizeF stringSize = g.MeasureString(str, font, bitmap.Width, sf);
			g.Dispose();
			return new Rectangle(point, new Size((int)stringSize.Width, (int)stringSize.Height));
		}

		public Bitmap GetRectangleBitmap(Rectangle rect)
		{
			Bitmap bmp = Bitmap;

			Bitmap bitmap = new Bitmap(rect.Width, rect.Height);
			for(int x = 0; x < rect.Width; x++)
			{
				if(x + rect.X < bmp.Width - 1)
					continue;

				for(int y = 0; y < rect.Height; y++)
				{
					if(y + rect.Y < bmp.Height - 1)
						continue;

					Color color = bmp.GetPixel(x + rect.X, y + rect.Y);
					bitmap.SetPixel(x, y, color);
				}
			}
			return bitmap;
		}

		public void CopySelection()
		{
			if(BitmapRectangle == null)
				return;

			Clipboard.SetImage(BitmapRectangle.Bitmap);
		}

		public void CutSelection()
		{
			if(BitmapRectangle == null)
				return;

			Clipboard.SetImage(BitmapRectangle.Bitmap);
			Rectangle rect = BitmapRectangle.Rectangle;

			// どこも選択されていない状態にする
			int x = BitmapRectangle.X;
			int y = BitmapRectangle.Y;

			BitmapRectangle.Dispose();
			BitmapRectangle = null;
			startPoint = new Point(-1, -1);
			endPoint = new Point(-1, -1);

			if(blankBitmap == null && (blankColorPoints == null || blankColorPoints.Count == 0))
			{
				ShowBitmap(Bitmap);
				return;
			}

			if(blankBitmap != null && (x != blankBitmap.Rectangle.X || y != blankBitmap.Rectangle.Y))
			{
				Graphics g1 = Graphics.FromImage(Bitmap);
				g1.DrawImage(blankBitmap.Bitmap, blankBitmap.Location);
				g1.Dispose();
				blankBitmap.Dispose();
				blankBitmap = null;
				ShowBitmap(Bitmap);
				return;
			}

			if(blankColorPoints != null && blankColorPoints.Count > 0)
			{
				int xMin = blankColorPoints.Min(x1 => x1.X);
				int yMin = blankColorPoints.Min(x1 => x1.Y);
				if(x != xMin || y != yMin)
				{
					Graphics g1 = Graphics.FromImage(Bitmap);
					foreach(var cp in blankColorPoints)
						Bitmap.SetPixel(cp.X, cp.Y, cp.Color);
					g1.Dispose();
					blankColorPoints.Clear();
					ShowBitmap(Bitmap);
					return;
				}
			}

			Bitmap whiteBitmap = new Bitmap(rect.Width, rect.Height);
			Graphics g = Graphics.FromImage(whiteBitmap);
			g.Clear(Color.White);
			g.Dispose();

			Bitmap newBitmap = new Bitmap(Bitmap);

			g = Graphics.FromImage(newBitmap);
			g.DrawImage(whiteBitmap, new Rectangle(rect.X, rect.Y, rect.Width, rect.Height), new Rectangle(0, 0, whiteBitmap.Width, whiteBitmap.Height), GraphicsUnit.Pixel);
			g.Dispose();
			whiteBitmap.Dispose();
			Bitmap = newBitmap;

			ShowBitmap(Bitmap);

			if(blankBitmap != null)
				blankBitmap.Dispose();
			blankBitmap = null;
		}


		private void UserControlImage_Resize(object sender, EventArgs e)
		{
			if(Bitmap == null)
				return;

			this.Invalidate(true);
			//pictureBox1.Invalidate();

			//if(pictureBox1.Width != 0)
			//	ShowBitmap1(Bitmap);

		}

		Brush GetBoderBrush()
		{
			return new System.Drawing.Drawing2D.HatchBrush(
				System.Drawing.Drawing2D.HatchStyle.BackwardDiagonal,
				Color.Black,
				Color.FromArgb(255, 128, 128, 128));
		}


		// 選択されている部分の境界線を描画する
		Bitmap DrawBoderRectangle(Bitmap bitmap)
		{
			Rectangle rect;
			if(BitmapRectangle == null)
			{
				rect = GetRectangle(startPoint, endPoint);
			}
			else
			{
				rect = BitmapRectangle.Rectangle;
			}

			Bitmap newBitmap = new Bitmap(bitmap);
			Graphics g = Graphics.FromImage(newBitmap);
			g.DrawRectangle(new Pen(GetBoderBrush()), rect);

			List<Rectangle> griperRects = new List<Rectangle>();
			griperRects.Add(new Rectangle(new Point(rect.X, rect.Y), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.X + rect.Width / 2 - GriperSize / 2, rect.Y), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.Right - GriperSize, rect.Y), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.X, rect.Bottom - GriperSize), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.X + rect.Width / 2 - GriperSize / 2, rect.Bottom - GriperSize), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.Right - GriperSize, rect.Bottom - GriperSize), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.X, rect.Y + rect.Height / 2 - GriperSize / 2), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.Right - GriperSize, rect.Y + rect.Height / 2 - GriperSize / 2), new Size(GriperSize, GriperSize)));
			g.FillRectangles(GetBoderBrush(), griperRects.ToArray());

			g.Dispose();

			return newBitmap;
		}


		public BitmapRectangle BitmapRectangle
		{
			set;
			get;
		} = null;

		private void VScrollBar1_Scroll(object sender, ScrollEventArgs e)
		{
			InvalidatePictureBox();
		}

		private void HScrollBar1_Scroll(object sender, ScrollEventArgs e)
		{
			InvalidatePictureBox();
		}


		Bitmap _bitmap = null;
		public Bitmap Bitmap
		{
			get
			{
				return _bitmap;
			}
			set
			{
				linePoints = null;

				if(_bitmap != value && _bitmap != null)
				{
					if(IsNewBitmap(_bitmap, value))
					{
						UndoBufs.InsertUndoBuf(new UndoBuf(_bitmap, DragDropRectangle));
						UndoBufs.ClearRedoBufs();
					}
				}

				_bitmap = value;
				pictureBox1.Invalidate();

				BitmapChanged?.Invoke(this, new EventArgs());
			}
		}

		bool IsNewBitmap(Bitmap oldBitmap, Bitmap newBitmap)
		{
			if(oldBitmap == null && newBitmap != null)
				return true;

			if(oldBitmap.Size != newBitmap.Size)
				return true;

			int width = oldBitmap.Size.Width;
			int height = oldBitmap.Size.Height;

			for(int x = 0; x< width; x++)
			{
				for(int y = 0; y < height; y++)
				{
					int oldArgb = oldBitmap.GetPixel(x, y).ToArgb();
					int newArgb = newBitmap.GetPixel(x, y).ToArgb();
					if(oldArgb != newArgb)
						return true;
				}
			}
			return false;
		}

		public event EventHandler BitmapChanged;
		UndoBufs UndoBufs = new UndoBufs();

		public void SizeChange(int newWidth, int newHeight)
		{
			UniteBitmapRectangle(new Point());

			Bitmap newBmp = new Bitmap(newWidth, newHeight);
			Graphics g = Graphics.FromImage(newBmp);
			g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
			g.DrawImage(Bitmap, 0, 0, newWidth, newHeight);
			g.Dispose();
			Bitmap = newBmp;

			ShowBitmap(Bitmap);
		}

		public bool RotateFlip(RotateFlipType rotateFlipType)
		{
			if(Bitmap == null)
				return false;

			if(BitmapRectangle != null)
				UniteBitmapRectangle(new Point());

			// bmpとは別にnewBitmapを作成する
			Bitmap newBitmap = new Bitmap(Bitmap);

			newBitmap.RotateFlip(rotateFlipType);
			Bitmap = newBitmap;
			ShowBitmap(Bitmap);

			// 左右上下反転と回転(後半)で追加
			if(BitmapRectangle != null)
			{
				BitmapRectangle.Dispose();
				BitmapRectangle = null;
			}

			return true;
		}

		public bool SelectionRotateFlip(RotateFlipType rotateFlipType)
		{
			if(Bitmap == null)
				return false;

			if(BitmapRectangle == null)
				return false;

			Bitmap bitmap = BitmapRectangle.Bitmap;
			Rectangle rect = BitmapRectangle.Rectangle;

			bitmap.RotateFlip(rotateFlipType);
			Bitmap bitmap1 = new Bitmap(bitmap);

			// Bitmapとは別にnewBitmapを作成する
			Bitmap newBitmap = new Bitmap(Bitmap);
			Graphics g = Graphics.FromImage(newBitmap);

			if(blankBitmap != null)
				g.DrawImage(blankBitmap.Bitmap, blankBitmap.Location);

			if(blankColorPoints != null)
			{
				foreach(var cp in blankColorPoints)
					bitmap1.SetPixel(cp.X, cp.Y, cp.Color);
			}

			Rectangle srcRect = new Rectangle(0, 0, bitmap1.Width, bitmap1.Height);
			g.DrawImage(bitmap1, rect, srcRect, GraphicsUnit.Pixel);
			g.Dispose();
			bitmap1.Dispose();

			ShowBitmap(DrawBoderRectangle(newBitmap));
			return true;
		}

		public bool LoadImageFile(string filePath)
		{
			Image image;
			try
			{
				image = Image.FromFile(filePath);
			}
			catch
			{
				return false;
			}

			if(Bitmap != null)
			{
				Bitmap.Dispose();
				_bitmap = null;
			}

			// ファイルがロックされないようにする
			Bitmap = new Bitmap(image);
			image.Dispose();

			Init();

			ShowBitmap(Bitmap);
			return true;
		}

		public void Init()
		{
			// スクロールに関する値をリセット
			ClearScrollInfo();

			ClearBitmapRectangle();
			ClearUndoBufs();

			if(borderPoints != null)
				borderPoints.Clear();

			if(blankBitmap != null)
				blankBitmap.Dispose();

			if(blankColorPoints != null)
				blankColorPoints.Clear();

			EditMode = EditMode.Selection;
		}

		public bool IsSelected()
		{
			return BitmapRectangle != null;
		}

		public void ShowBitmap(Bitmap bitmap)
		{
			OptimizationScrloolBar(bitmap);
			InvalidatePictureBox();
		}

		public void ShowBitmap1(Bitmap bitmap)
		{
			OptimizationScrloolBar(bitmap);

			// スクロールバー位置より元画像をトリミングして表示する
			Bitmap bitmap1 = GetBitmapDrawPictureBox(bitmap);

			if(BitmapRectangle == null)
			{
				Graphics g = Graphics.FromImage(bitmap1);
				int x = bitmap.Width - ScrollBarPosX - GriperSize;
				int y = bitmap.Height - ScrollBarPosY - GriperSize;
				g.DrawRectangle(new Pen(Color.Red), new Rectangle(x, y, GriperSize, GriperSize));

				x = bitmap.Width / 2 - ScrollBarPosX - GriperSize / 2;
				y = bitmap.Height - ScrollBarPosY - GriperSize;
				g.DrawRectangle(new Pen(Color.Black), new Rectangle(x, y, GriperSize, GriperSize));

				x = bitmap.Width - ScrollBarPosX - GriperSize;
				y = bitmap.Height / 2 - ScrollBarPosY - GriperSize / 2;
				g.DrawRectangle(new Pen(Color.Black), new Rectangle(x, y, GriperSize, GriperSize));

				g.DrawRectangle(new Pen(GetBoderBrush()), new Rectangle(-1, -1, bitmap.Width + 1, bitmap.Height + 1));

				g.Dispose();
			}

			pictureBox1.Image = bitmap1;
		}

		public void Undo()
		{
			if(BitmapRectangle != null)
			{
				ClearBitmapRectangle();
				pictureBox1.Invalidate();
				return;
			}

			if(UndoBufs.CanUndo())
			{
				if(_bitmap != null)
					UndoBufs.InsertRedoBuf(new UndoBuf(_bitmap, DragDropRectangle));

				UndoBuf undoBuf = UndoBufs.RemoveUndoBuf();

				_bitmap = undoBuf.Bitmap;

				startPoint = new Point(-1, -1);
				endPoint = new Point(-1, -1);

				if(_bitmap != null)
				{
					OptimizationScrloolBar(Bitmap);
					pictureBox1.Invalidate();
				}
				BitmapChanged?.Invoke(this, new EventArgs());
			}
		}

		public void Redo()
		{
			if(BitmapRectangle != null)
			{
				ClearBitmapRectangle();
				pictureBox1.Invalidate();
				return;
			}

			if(UndoBufs.CanRedo())
			{
				UndoBufs.InsertUndoBuf(new UndoBuf(_bitmap, DragDropRectangle));

				UndoBuf redoBuf = UndoBufs.RemoveRedoBuf();

				_bitmap = redoBuf.Bitmap;

				startPoint = new Point(-1, -1);
				endPoint = new Point(-1, -1);

				if(_bitmap != null)
				{
					OptimizationScrloolBar(Bitmap);
					pictureBox1.Invalidate();
				}

				BitmapChanged?.Invoke(this, new EventArgs());
			}
		}

		public int UndoBufCount
		{
			get { return UndoBufs.UndoBufsCount; }
		}
		public int RedoBufCount
		{
			get { return UndoBufs.RedoBufsCount; }
		}

		public bool CanUndo()
		{
			return UndoBufs.CanUndo();
		}

		public bool CanRedo()
		{
			return UndoBufs.CanRedo();
		}

		public void ClearUndoBufs()
		{
			UndoBufs.ClearUndoBufs();
			BitmapChanged?.Invoke(this, new EventArgs());
		}

		public int ScrollBarPosX
		{
			get { return hScrollBar1.Value; }
			set { hScrollBar1.Value = value; }
		}

		public int ScrollBarPosY
		{
			get { return vScrollBar1.Value; }
			set { vScrollBar1.Value = value; }
		}

		public Rectangle DragDropRectangle
		{
			get;
			private set;
		}  = new Rectangle();

		Point startPoint = new Point(-1, -1);
		Point endPoint = new Point(-1, -1);

		Rectangle GetRectangle(Point pt1, Point pt2)
		{
			if(pt1.X < 0 || pt2.X < 0)
				return new Rectangle();

			if(pt1.X < pt2.X)
			{
				if(pt1.Y < pt2.Y)
					return new Rectangle(pt1, new Size(pt2.X - pt1.X, pt2.Y - pt1.Y));
				else
					return new Rectangle(new Point(pt1.X, pt2.Y), new Size(pt2.X - pt1.X, -(pt2.Y - pt1.Y)));
			}
			else
			{
				if(pt1.Y < pt2.Y)
					return new Rectangle(new Point(pt2.X, pt1.Y), new Size(pt1.X - pt2.X, pt2.Y - pt1.Y));
				else
					return new Rectangle(pt2, new Size(pt1.X - pt2.X, pt1.Y - pt2.Y));
			}
		}

		bool isMouseDown = false;

		public class PictureBox1MouseDownArgs : EventArgs
		{
			public PictureBox1MouseDownArgs(int x, int y, Color color)
			{
				X = x;
				Y = y;
				Color = color;
			}
			public int X
			{
				get;
				private set;
			}
			public int Y
			{
				get;
				private set;
			}
			public Color Color
			{
				get;
				private set;
			}
		}

		public delegate void PictureBoxMouseDownHandler(object sender, PictureBox1MouseDownArgs e);
		public event PictureBoxMouseDownHandler PictureBoxMouseDown;

		List<Point> linePoints = null;

		bool IsPointInRectangle(Point pt, Rectangle rect)
		{
			if(pt.X < rect.Left)
				return false;
			if(pt.Y < rect.Top)
				return false;
			if(rect.Right < pt.X)
				return false;
			if(rect.Bottom < pt.Y)
				return false;

			return true;
		}

		bool isMouseDownForMove = false;
		bool isMouseDownForSizeChange = false;

		// 移動前の座標
		int rectangleMoveStartLocationX = 0;
		int rectangleMoveStartLocationY = 0;

		int rectangleMoveStartMousePosX = 0;
		int rectangleMoveStartMousePosY = 0;

		Griper dragGriper = Griper.Out;

		void OnMouseDownInBitmapRectangle(Point bitmapPoint)
		{
			dragGriper = IsMouseOnGriper(new Point(bitmapPoint.X, bitmapPoint.Y));
			if(dragGriper != Griper.In && dragGriper != Griper.Out)
			{
				OnMouseDownForSizeChangeSelection(bitmapPoint, dragGriper);
			}
			else if(IsMouseOnGriper(new Point(bitmapPoint.X, bitmapPoint.Y)) == Griper.In)
			{
				OnMouseDownForMoveSelection(bitmapPoint);
			}
		}

		void OnMouseDownForSizeChangeSelection(Point clickedBitmapPoint, Griper griper)
		{
			isMouseDownForSizeChange = true;
			pictureBox1.Capture = true;

			rectangleMoveStartMousePosX = clickedBitmapPoint.X;
			rectangleMoveStartMousePosY = clickedBitmapPoint.Y;

			rectangleMoveStartLocationX = BitmapRectangle.X;
			rectangleMoveStartLocationY = BitmapRectangle.Y;

			dragGriper = griper;
		}

		public void ClearBitmapRectangle(bool doesUnite)
		{
			if(doesUnite)
			{
				UniteBitmapRectangle(new Point());
				ShowBitmap(Bitmap);
			}
			else
				ClearBitmapRectangle();

			isMouseDown = false;
			isMouseDownForMove = false;
			isMouseDownForSizeChange = false;
		}

		public void ClearBitmapRectangle()
		{
			if(BitmapRectangle != null)
				BitmapRectangle.Dispose();

			if(blankBitmap != null)
			{
				blankBitmap.Dispose();
				blankBitmap = null;
			}
			BitmapRectangle = null;

			isMouseDown = false;
			isMouseDownForMove = false;
			isMouseDownForSizeChange = false;
		}


		void OnMouseMoveForMoveFreeLineSelection(Point bitmapPoint)
		{
			if(BitmapRectangle == null)
				return;

			int xMove = bitmapPoint.X - rectangleMoveStartMousePosX;
			int yMove = bitmapPoint.Y - rectangleMoveStartMousePosY;
			BitmapRectangle.X = rectangleMoveStartLocationX + xMove;
			BitmapRectangle.Y = rectangleMoveStartLocationY + yMove;

			InvalidatePictureBox();
		}

		void OnMouseUpForMoveFreeLineSelection(Point bitmapPoint)
		{
			isMouseDownForMove = false;
			if(BitmapRectangle != null)
				BitmapRectangle.UpdatePosition();
		}

		void OnMouseMoveForSizeChangeSelection(Point bitmapPoint)
		{
			int xMove = bitmapPoint.X - rectangleMoveStartMousePosX;
			int yMove = bitmapPoint.Y - rectangleMoveStartMousePosY;

			int x, y, width, height;
			x = BitmapRectangle.X;
			y = BitmapRectangle.Y;
			width = BitmapRectangle.Width;
			height = BitmapRectangle.Height;

			bool isRatioMaintained = false;
			if(Control.ModifierKeys.HasFlag(Keys.Control))
				isRatioMaintained = true;

			if(dragGriper == Griper.East)
			{
				width = BitmapRectangle.OldWidth + xMove;
				if(isRatioMaintained)
					height = BitmapRectangle.OldHeight * width / BitmapRectangle.OldWidth;
			}
			if(dragGriper == Griper.South)
			{
				height = BitmapRectangle.OldHeight + yMove;
				if(isRatioMaintained)
					width = BitmapRectangle.OldWidth * height / BitmapRectangle.OldHeight;
			}
			if(dragGriper == Griper.SouthEast)
			{
				width = BitmapRectangle.OldWidth + xMove;
				height = BitmapRectangle.OldHeight + yMove;
				if(isRatioMaintained)
					width = BitmapRectangle.OldWidth * height / BitmapRectangle.OldHeight;
			}
			if(dragGriper == Griper.North)
			{
				y = BitmapRectangle.OldY + yMove;
				height = BitmapRectangle.OldHeight - yMove;
				if(isRatioMaintained)
					width = BitmapRectangle.OldWidth * height / BitmapRectangle.OldHeight;
			}
			if(dragGriper == Griper.West)
			{
				x = BitmapRectangle.OldX + xMove;
				width = BitmapRectangle.OldWidth - xMove;
				if(isRatioMaintained)
					height = BitmapRectangle.OldHeight * width / BitmapRectangle.OldWidth;
			}
			if(dragGriper == Griper.NorthWest)
			{
				x = BitmapRectangle.OldX + xMove;
				width = BitmapRectangle.OldWidth - xMove;
				y = BitmapRectangle.OldY + yMove;
				height = BitmapRectangle.OldHeight - yMove;
				if(isRatioMaintained)
				{
					width = BitmapRectangle.OldWidth * height / BitmapRectangle.OldHeight;
					x = BitmapRectangle.OldX - (width - BitmapRectangle.OldWidth);
				}
			}
			if(dragGriper == Griper.NorthEast)
			{
				width = BitmapRectangle.OldWidth + xMove;
				y = BitmapRectangle.OldY + yMove;
				height = BitmapRectangle.OldHeight - yMove;
				if(isRatioMaintained)
				{
					width = BitmapRectangle.OldWidth * height / BitmapRectangle.OldHeight;
				}
			}
			if(dragGriper == Griper.SouthWest)
			{
				x = BitmapRectangle.OldX + xMove;
				width = BitmapRectangle.OldWidth - xMove;
				height = BitmapRectangle.OldHeight + yMove;
				if(isRatioMaintained)
				{
					width = BitmapRectangle.OldWidth * height / BitmapRectangle.OldHeight;
					x = BitmapRectangle.OldX - (width - BitmapRectangle.OldWidth);
				}
			}

			if(width <= 1 || height <= 1)
			{
				isMouseDownForSizeChange = false;
				startPoint = new Point(BitmapRectangle.Rectangle.Left, BitmapRectangle.Rectangle.Top);
				endPoint = new Point(BitmapRectangle.Rectangle.Right, BitmapRectangle.Rectangle.Bottom);
				BitmapRectangle.UpdatePosition();
				return;
			}

			BitmapRectangle.X = x;
			BitmapRectangle.Width = width;
			BitmapRectangle.Y = y;
			BitmapRectangle.Height = height;

			InvalidatePictureBox();
		}

		public bool ReplaceColor(Color oldColor, Color newColor, bool isOldColorTransparent, bool isNewColorTransparent)
		{
			if(BitmapRectangle != null)
				UniteBitmapRectangle(new Point());

			Bitmap bmp = Bitmap;

			if(isOldColorTransparent)
				oldColor = Color.FromArgb(0, 0, 0, 0);
			if(isNewColorTransparent)
				newColor = Color.FromArgb(0, 0, 0, 0);

			Bitmap = GetBitmapReplaceColor(bmp, oldColor, newColor);
			ShowBitmap(Bitmap);

			return true;
		}

		Bitmap GetBitmapReplaceColor(Bitmap sourceBmp, Color beforeColor, Color afterColor)
		{
			int width = sourceBmp.Width;
			int height = sourceBmp.Height;
			Bitmap newBitmap = new Bitmap(width, height);

			for(int x = 0; x < width; x++)
			{
				for(int y = 0; y < height; y++)
				{
					Color color = sourceBmp.GetPixel(x, y);

					// 透明と黒を混同しないようにする
					if(beforeColor != Color.FromArgb(0, 0, 0, 0) && color != Color.FromArgb(0, 0, 0, 0))
						newBitmap.SetPixel(x, y, afterColor);
					else if(beforeColor == Color.FromArgb(0, 0, 0, 0) && color == Color.FromArgb(0, 0, 0, 0))
						newBitmap.SetPixel(x, y, afterColor);
					else
						newBitmap.SetPixel(x, y, color);
				}
			}
			return newBitmap;
		}

		Bitmap GetBitmapReplaceColor(Bitmap sourceBmp, Color beforeColor, Color afterColor, int differenceR, int differenceG, int differenceB)
		{
			int width = sourceBmp.Width;
			int height = sourceBmp.Height;
			Bitmap newBitmap = new Bitmap(width, height);

			int minR = beforeColor.R - differenceR;
			int maxR = beforeColor.R + differenceR;
			int minG = beforeColor.G - differenceG;
			int maxG = beforeColor.G + differenceG;
			int minB = beforeColor.B - differenceB;
			int maxB = beforeColor.B + differenceB;

			if(minR < 0)
				minR = 0;
			if(maxR > 255)
				maxR = 255;
			if(minG < 0)
				minG = 0;
			if(maxG > 255)
				maxG = 255;
			if(minB < 0)
				minB = 0;
			if(maxB > 255)
				maxB = 255;

			for(int x = 0; x < width; x++)
			{
				for(int y = 0; y < height; y++)
				{
					Color color = sourceBmp.GetPixel(x, y);

					// 透明と黒を混同しないようにする
					if(beforeColor != Color.FromArgb(0, 0, 0, 0) && color != Color.FromArgb(0, 0, 0, 0) && minR <= color.R && color.R <= maxR && minG <= color.G && color.G <= maxG && minB <= color.B && color.B <= maxB)
						newBitmap.SetPixel(x, y, afterColor);
					else if(beforeColor == Color.FromArgb(0, 0, 0, 0) && color == Color.FromArgb(0, 0, 0, 0))
						newBitmap.SetPixel(x, y, afterColor);
					else
						newBitmap.SetPixel(x, y, color);
				}
			}
			return newBitmap;
		}

		public bool ReplaceColorInRectangle(Color oldColor, Color newColor, bool isOldColorTransparent, bool isNewColorTransparent,
			int differenceR, int differenceG, int differenceB)
		{
			if(BitmapRectangle == null)
				return false;

			if(isOldColorTransparent)
				oldColor = Color.FromArgb(0, 0, 0, 0);
			if(isNewColorTransparent)
				newColor = Color.FromArgb(0, 0, 0, 0);

			Bitmap oldSelectedBitmap = BitmapRectangle.Bitmap;
			Rectangle rect = BitmapRectangle.Rectangle;

			Bitmap newSelectedBitmap;
			Bitmap newBitmap;

			if(isNewColorTransparent)
			{
				newBitmap = SetSelectionTransparent(rect, oldColor, differenceR, differenceG, differenceB);
			}
			else
			{
				if(!isOldColorTransparent)
					newSelectedBitmap = GetBitmapReplaceColor(oldSelectedBitmap, oldColor, newColor, differenceR, differenceG, differenceB);
				else
					newSelectedBitmap = GetBitmapReplaceColor(oldSelectedBitmap, oldColor, newColor, 0, 0, 0);

				BitmapRectangle.Bitmap = newSelectedBitmap;
				newBitmap = new Bitmap(Bitmap);
				Graphics g = Graphics.FromImage(newBitmap);

				if(blankBitmap != null)
					g.DrawImage(blankBitmap.Bitmap, blankBitmap.Location);

				if(blankColorPoints != null)
				{
					foreach(var cp in blankColorPoints)
						newBitmap.SetPixel(cp.X, cp.Y, cp.Color);
				}

				Rectangle srcRect = new Rectangle(0, 0, rect.Width, rect.Height);
				g.DrawImage(newSelectedBitmap, rect, srcRect, GraphicsUnit.Pixel);
				g.Dispose();
			}

			ShowBitmap(DrawBoderRectangle(newBitmap));

			return true;
		}

		public Bitmap SetSelectionTransparent(Rectangle rect, Color beforeColor)
		{
			Bitmap bmp = Bitmap;

			int width = bmp.Width;
			int height = bmp.Height;

			Bitmap newBmp = new Bitmap(width, height);
			Bitmap newRectBmp = new Bitmap(rect.Width, rect.Height);

			for(int x = 0; x < width; x++)
			{
				for(int y = 0; y < height; y++)
				{
					Color color = bmp.GetPixel(x, y);

					if((x < rect.Left || rect.Right <= x) || (y < rect.Top || rect.Bottom <= y))
					{
						newBmp.SetPixel(x, y, color);
					}
					else
					{
						if(beforeColor.ToArgb() == color.ToArgb())
						{
							newBmp.SetPixel(x, y, Color.FromArgb(0, 0, 0, 0));
							newRectBmp.SetPixel(x - rect.X, y - rect.Y, Color.FromArgb(0, 0, 0, 0));
						}
						else
							newBmp.SetPixel(x, y, color);
					}
				}
			}
			BitmapRectangle.Bitmap = newRectBmp;
			return newBmp;
		}

		public Bitmap SetSelectionTransparent(Rectangle rect, Color beforeColor, int differenceR, int differenceG, int differenceB)
		{
			int minR = beforeColor.R - differenceR;
			int maxR = beforeColor.R + differenceR;
			int minG = beforeColor.G - differenceG;
			int maxG = beforeColor.G + differenceG;
			int minB = beforeColor.B - differenceB;
			int maxB = beforeColor.B + differenceB;

			if(minR < 0)
				minR = 0;
			if(maxR > 255)
				maxR = 255;
			if(minG < 0)
				minG = 0;
			if(maxG > 255)
				maxG = 255;
			if(minB < 0)
				minB = 0;
			if(maxB > 255)
				maxB = 255;

			for(int x = 0; x < BitmapRectangle.Bitmap.Width; x++)
			{
				for(int y = 0; y < BitmapRectangle.Bitmap.Height; y++)
				{
					Color color = BitmapRectangle.Bitmap.GetPixel(x, y);

					if(minR <= color.R && color.R <= maxR && minG <= color.G && color.G <= maxG && minB <= color.B && color.B <= maxB)
					{

						BitmapRectangle.Bitmap.SetPixel(x, y, Color.FromArgb(0, 0, 0, 0));
					}
					else
					{
						BitmapRectangle.Bitmap.SetPixel(x, y, color);
					}
				}
			}

			Bitmap newBitmap = new Bitmap(Bitmap);
			Graphics g = Graphics.FromImage(newBitmap);
			if(blankBitmap != null)
				g.DrawImage(blankBitmap.Bitmap, blankBitmap.Location);

			if(blankColorPoints != null)
			{
				foreach(var cp in blankColorPoints)
					newBitmap.SetPixel(cp.X, cp.Y, cp.Color);
			}

			g.DrawImage(BitmapRectangle.Bitmap, BitmapRectangle.Location);
			g.Dispose();

			return newBitmap;
		}

		public bool ReplaceColor(Color oldColor, Color newColor, bool isOldColorTransparent, bool isNewColorTransparent,
			int differenceR, int differenceG, int differenceB)
		{
			if(BitmapRectangle != null)
				UniteBitmapRectangle(new Point());

			Bitmap bmp = Bitmap;

			if(isOldColorTransparent)
				oldColor = Color.FromArgb(0, 0, 0, 0);
			if(isNewColorTransparent)
				newColor = Color.FromArgb(0, 0, 0, 0);

			Bitmap newBitmap;
			if(!isOldColorTransparent)
				newBitmap = GetBitmapReplaceColor(bmp, oldColor, newColor, differenceR, differenceG, differenceB);
			else
				newBitmap = GetBitmapReplaceColor(bmp, oldColor, newColor, 0, 0, 0);

			Bitmap = newBitmap;
			ShowBitmap(Bitmap);
			return true;
		}

		public bool ReverseColor()
		{
			if(Bitmap == null)
				return false;

			if(BitmapRectangle != null)
				UniteBitmapRectangle(new Point());

			Bitmap bmp = Bitmap;

			int width = bmp.Width;
			int height = bmp.Height;

			Bitmap newBitmap = new Bitmap(width, height);

			for(int x = 0; x < width; x++)
			{
				for(int y = 0; y < height; y++)
				{
					Color color = bmp.GetPixel(x, y);
					int r = 255 - color.R;
					int g = 255 - color.G;
					int b = 255 - color.B;
					newBitmap.SetPixel(x, y, Color.FromArgb(r, g, b));
				}
			}
			Bitmap = newBitmap;
			ShowBitmap(Bitmap);

			return true;
		}

		public bool ReverseColorInRectangle()
		{
			if(Bitmap == null)
				return false;

			if(BitmapRectangle == null)
				return false;

			Bitmap bitmap = BitmapRectangle.Bitmap;
			Rectangle rect = BitmapRectangle.Rectangle;

			Bitmap newSelectBitmap = new Bitmap(BitmapRectangle.OldBitmap);
			int width = newSelectBitmap.Width;
			int height = newSelectBitmap.Height;
			for(int x = 0; x < width; x++)
			{
				for(int y = 0; y < height; y++)
				{
					Color color = newSelectBitmap.GetPixel(x, y);
					int red = 255 - color.R;
					int green = 255 - color.G;
					int blue = 255 - color.B;
					newSelectBitmap.SetPixel(x, y, Color.FromArgb(red, green, blue));
				}
			}
			BitmapRectangle = new BitmapRectangle(newSelectBitmap, rect);

			Bitmap newBitmap = new Bitmap(Bitmap);
			Graphics g = Graphics.FromImage(newBitmap);

			if(blankBitmap != null)
				g.DrawImage(blankBitmap.Bitmap, blankBitmap.Location);

			if(blankColorPoints != null)
			{
				foreach(var cp in blankColorPoints)
					newBitmap.SetPixel(cp.X, cp.Y, cp.Color);
			}

			Rectangle srcRect = new Rectangle(0, 0, BitmapRectangle.Width, BitmapRectangle.Height);
			g.DrawImage(BitmapRectangle.Bitmap, rect, srcRect, GraphicsUnit.Pixel);
			g.Dispose();

			InvalidatePictureBox();
			return true;
		}

		void UniteBitmapRectangle(Point mouseDownPoint)
		{
			Bitmap bitmap1 = new Bitmap(Bitmap);
			Graphics g = Graphics.FromImage(bitmap1);

			if(blankBitmap != null)
			{
				Point pt = blankBitmap.Location;

				g.DrawImage(blankBitmap.Bitmap, pt);
				blankBitmap.Dispose();
				blankBitmap = null;
			}

			if(BitmapRectangle != null)
			{
				Point pt = BitmapRectangle.Location;
				g.DrawImage(BitmapRectangle.Bitmap, pt);
				g.Dispose();

				BitmapRectangle.Dispose();
				BitmapRectangle = null;
			}

			Bitmap = bitmap1;

			startPoint = new Point(-1, -1);
			endPoint = new Point(-1, -1);
		}

		void OnMouseDownOutOfBitmap()
		{
			startPoint = new Point(-1, -1);
			ShowBitmap(Bitmap);
		}

		void OnMouseDownOutOfSelection(Point bitmapPoint)
		{
			startPoint = new Point(-1, -1);
			endPoint = new Point(-1, -1);

			UniteBitmapRectangle(bitmapPoint);
			ShowBitmap(Bitmap);

			Console.WriteLine("BitmapRectangle = null");
		}

		bool isMouseDownForFieldSizeChange = false;
		int mousePositionStartFieldSizeChangeX = 0;
		int mousePositionStartFieldSizeChangeY = 0;
		Griper griperField = Griper.Out;

		void OnMouseDownForSelectRectangle(Point bitmapPoint)
		{
			Griper griper = IsMouseOnFirldGriper(bitmapPoint);
			if(griper == Griper.South || griper == Griper.East || griper == Griper.SouthEast)
			{
				griperField = griper;
				isMouseDownForFieldSizeChange = true;

				mousePositionStartFieldSizeChangeX = bitmapPoint.X;
				mousePositionStartFieldSizeChangeY = bitmapPoint.Y;

				DrawFieldGriper(new Point(Bitmap.Width, Bitmap.Height));
			}
			else
				isMouseDown = true;

			pictureBox1.Capture = true;
			startPoint = new Point(bitmapPoint.X, bitmapPoint.Y);
		}

		void OnMouseDownForDrawRectangle(Point bitmapPoint)
		{
			isMouseDown = true;
			pictureBox1.Capture = true;
			startPoint = new Point(bitmapPoint.X, bitmapPoint.Y);
		}

		void OnMouseDownForDrawLine(Point bitmapPoint)
		{
			isMouseDown = true;
			pictureBox1.Capture = true;
			startPoint = new Point(bitmapPoint.X, bitmapPoint.Y);
		}

		void OnMouseDownForDrawFreeLine(Point bitmapPoint)
		{
			isMouseDown = true;
			pictureBox1.Capture = true;
			startPoint = new Point(bitmapPoint.X, bitmapPoint.Y);

			linePoints = new List<Point>();
			linePoints.Add(bitmapPoint);
		}

		void OnMouseDownForSelectFreeArea(Point bitmapPoint)
		{
			if(borderPoints != null && borderPoints.Count > 0)
			{
				OnMouseDownForMoveFreeLineSelection(new Point(bitmapPoint.X, bitmapPoint.Y));
				borderPoints.Clear();
				return;
			}

			isMouseDown = true;
			pictureBox1.Capture = true;
			startPoint = new Point(bitmapPoint.X, bitmapPoint.Y);

			linePoints = new List<Point>();
			linePoints.Add(bitmapPoint);
		}

		void OnMouseMoveForSelectFreeArea(Point bitmapPoint)
		{
			endPoint = bitmapPoint;
			linePoints.Add(bitmapPoint);

			InvalidatePictureBox();
		}

		public Color DrawColor
		{
			set;
			get;
		}
		 = Color.Black;

		public class ColorPoint
		{
			public ColorPoint(int x, int y, Color color)
			{
				X = x;
				Y = y;
				Color = color;
			}

			public int X
			{
				get;
				set;
			}
			= 0;

			public int Y
			{
				get;
				set;
			}
			= 0;

			public Color Color
			{
				get;
				set;
			}
		}

		void OnMouseDownForMoveFreeLineSelection(Point bitmapPoint)
		{
			Bitmap temp = new Bitmap(Bitmap.Width, Bitmap.Height);
			foreach(Point pt in borderPoints)
				temp.SetPixel(pt.X, pt.Y, Color.Black);

			FillColor(temp, bitmapPoint, Color.Red);

			List<ColorPoint> colorPoints = new List<ColorPoint>();
			for(int x = 0; x < Bitmap.Width; x++)
			{
				for(int y = 0; y < Bitmap.Height; y++)
				{
					Color color = temp.GetPixel(x, y);
					if(color.ToArgb() == Color.Red.ToArgb())
					{
						colorPoints.Add(new ColorPoint(x, y, Bitmap.GetPixel(x, y)));
					}
				}
			}

			int pCount = borderPoints.Count;
			borderPoints.Clear();

			// 自由曲線が閉じていない場合は全体が選択されてしまう。
			// この場合はなにもしないで処理を終了する
			if(colorPoints.Count == Bitmap.Width * Bitmap.Height - pCount)
			{
				colorPoints.Clear();
				ShowBitmap(Bitmap);
				return;
			}

			// 移動元は白（ただし透明部分は透明）にする
			if(colorPoints != null && colorPoints.Count > 0)
			{
				isMouseDownForMove = true;

				CreateBlankBitmap(colorPoints);
			}

			// 移動元になるビットマップを作成する
			int minX = colorPoints.Min(x => x.X);
			int maxX = colorPoints.Max(x => x.X);
			int minY = colorPoints.Min(x => x.Y);
			int maxY = colorPoints.Max(x => x.Y);

			Bitmap bitmap = new Bitmap(maxX - minX + 1, maxY - minY + 1);
			foreach(var cp in colorPoints)
			{
				bitmap.SetPixel(cp.X - minX, cp.Y - minY, cp.Color);
			}

			// 移動元になるビットマップが作成できたらBitmapRectangleにセットする
			BitmapRectangle = new BitmapRectangle(bitmap, new Rectangle(new Point(minX, minY), new Size(maxX - minX, maxY - minY)));
			this.rectangleMoveStartMousePosX = bitmapPoint.X;
			this.rectangleMoveStartMousePosY = bitmapPoint.Y;
			this.rectangleMoveStartLocationX = minX;
			this.rectangleMoveStartLocationY = minY;
		}

		List<ColorPoint> blankColorPoints = null;
		void CreateBlankBitmap(List<ColorPoint> pts)
		{
			blankColorPoints = new List<ColorPoint>();
			foreach(ColorPoint cp in pts)
			{
				Color color = cp.Color;
				if(color != Color.FromArgb(0, 0, 0, 0))
					color = Color.White;
				blankColorPoints.Add(new ColorPoint(cp.X, cp.Y, color));
			}
		}

		public void FillColor(Bitmap bitmap, Point bitmapPoint, Color newColor)
		{
			if(bitmap == null)
				return;

			Color oldColor = bitmap.GetPixel(bitmapPoint.X, bitmapPoint.Y);

			isCheckedFillColor = new bool[bitmap.Width, bitmap.Height];

			if(oldColor.ToArgb() == newColor.ToArgb())
				return;

			var list = FillLine(bitmap, bitmapPoint.X, bitmapPoint.Y, oldColor, newColor);
			List<Point> nextPoints = new List<Point>();

			while(true)
			{
				foreach(var pt in list)
					nextPoints.AddRange(FillLine(bitmap, pt.X, pt.Y, oldColor, newColor));
				list = nextPoints.ToList();
				nextPoints.Clear();

				if(list.Count == 0)
					break;
			}
		}

		// スクロールバー位置より元画像をトリミングして取得する
		Bitmap GetBitmapDrawPictureBox(Bitmap bitmap)
		{
			Bitmap newBitmap = new Bitmap(pictureBox1.Width, pictureBox1.Height);
			RectangleF rectSrc = new RectangleF(
				hScrollBar1.Value,
				vScrollBar1.Value,
				pictureBox1.Width / DisplayMagnification,
				pictureBox1.Height / DisplayMagnification);

			Graphics g = Graphics.FromImage(newBitmap);
			Rectangle destRect = new Rectangle(0, 0, pictureBox1.Width, pictureBox1.Height);
			g.DrawImage(bitmap, destRect, rectSrc, GraphicsUnit.Pixel);
			g.Dispose();

			return newBitmap;
		}

		// スクロールバー位置より元画像をトリミングして取得する
		Bitmap GetBitmapDrawPictureBox_old(Bitmap bitmap)
		{
			Bitmap newBitmap = new Bitmap(pictureBox1.Width, pictureBox1.Height);
			RectangleF rectSrc = new RectangleF(
				this.hScrollBar1.Value,
				this.vScrollBar1.Value,
				this.pictureBox1.Width,
				this.pictureBox1.Height);

			Graphics g = Graphics.FromImage(newBitmap);
			g.DrawImage(bitmap, 0, 0, rectSrc, GraphicsUnit.Pixel);
			g.Dispose();

			return newBitmap;
		}

		public class PictureBox1MouseUpArgs : EventArgs
		{
			public PictureBox1MouseUpArgs(Rectangle rectangle)
			{
				Rectangle = rectangle;
			}
			public PictureBox1MouseUpArgs(List<Point> points)
			{
				Points = points;
			}
			public Rectangle Rectangle
			{
				get;
				private set;
			}
			public List<Point> Points
			{
				get;
				private set;
			}
		}

		public delegate void PictureBoxMouseUpHandler(object sender, PictureBox1MouseUpArgs e);

		void SelectRectangle()
		{
			isMouseDown = false;

			Rectangle rect = GetRectangle(startPoint, endPoint);
			if(rect.IsEmpty || rect.Width == 0 || rect.Height == 0)
			{
				startPoint = new Point(-1, -1);
				endPoint = new Point(-1, -1);
				return;
			}

			Bitmap bitmap = new Bitmap(rect.Width, rect.Height);
			Graphics g = Graphics.FromImage(bitmap);
			Rectangle destRect = new Rectangle(0, 0, rect.Width, rect.Height);
			g.DrawImage(Bitmap, destRect, rect, GraphicsUnit.Pixel);
			g.Dispose();

			BitmapRectangle = new BitmapRectangle(new Bitmap(bitmap), rect);
			g.Dispose();

			if(blankBitmap == null)
				CreateBlankBitmap(BitmapRectangle.Rectangle);
		}

		void OnMouseUpForSelectRectangle()
		{
			SelectRectangle();
		}

		void OnMouseUpForDrawRectangle()
		{
			isMouseDown = false;

			Rectangle rect = GetRectangle(startPoint, endPoint);
			if(rect.IsEmpty || rect.Width == 0 || rect.Height == 0)
				return;

			Bitmap rectBitmap = new Bitmap(rect.Width + 1, rect.Height + 1);
			Graphics g = Graphics.FromImage(rectBitmap);
			if(EditMode == EditMode.DrawRectangle)
				g.DrawRectangle(new Pen(DrawColor), new Rectangle(0,0,rect.Width, rect.Height));
			if(EditMode == EditMode.FillRectangle)
				g.FillRectangle(new SolidBrush(DrawColor), new Rectangle(0, 0, rect.Width, rect.Height));
			if(EditMode == EditMode.DrawEllipse)
				g.DrawEllipse(new Pen(DrawColor), new Rectangle(0, 0, rect.Width, rect.Height));
			if(EditMode == EditMode.FillEllipse)
				g.FillEllipse(new SolidBrush(DrawColor), new Rectangle(0, 0, rect.Width, rect.Height));
			g.Dispose();

			BitmapRectangle = new BitmapRectangle(new Bitmap(rectBitmap), rect);
		}

		void OnMouseUpForDrawLine(Point bitmapPoint)
		{
			isMouseDown = false;

			endPoint = bitmapPoint;
			Bitmap newBitmap = new Bitmap(Bitmap);

			Graphics g = Graphics.FromImage(newBitmap);
			if(EditMode == EditMode.DrawLine)
			{
				int x1 = startPoint.X;
				int y1 = startPoint.Y;
				int x2 = endPoint.X;
				int y2 = endPoint.Y;

				g.DrawLine(new Pen(DrawColor), x1, y1, x2, y2);
			}

			g.Dispose();

			Bitmap = newBitmap;
			ShowBitmap(newBitmap);

			startPoint = new Point(-1, -1);
			endPoint = new Point(-1, -1);
		}

		void OnMouseUpForDrawFreeLine(Point bitmapPoint)
		{
			isMouseDown = false;

			endPoint = bitmapPoint;
			Bitmap newBitmap = new Bitmap(Bitmap);

			linePoints.Add(bitmapPoint);

			if(linePoints.Count < 2)
				return;

			Point fromPoint = linePoints[0];
			Point[] points = linePoints.Skip(1).ToArray();
			Graphics g = Graphics.FromImage(newBitmap);
			if(EditMode == EditMode.FreeLine || EditMode == EditMode.SelectionFree)
			{
				foreach(Point toPoint in points)
				{
					g.DrawLine(new Pen(DrawColor), fromPoint.X, fromPoint.Y, toPoint.X, toPoint.Y);
					fromPoint = toPoint;
				}
			}

			g.Dispose();

			Bitmap = newBitmap;
			ShowBitmap(newBitmap);

			startPoint = new Point(-1, -1);
			endPoint = new Point(-1, -1);
		}

		void DrawFreeLine(Bitmap bitmap, List<Point> points, Pen pen)
		{
			if(points.Count < 2)
				return;

			Point fromPoint = points[0];
			Point[] points1 = points.ToArray();
			Graphics g = Graphics.FromImage(bitmap);
			if(EditMode == EditMode.SelectionFree)
			{
				foreach(Point toPoint in points1)
				{
					g.DrawLine(pen, fromPoint.X, fromPoint.Y, toPoint.X, toPoint.Y);
					fromPoint = toPoint;
				}
			}
			g.Dispose();
		}

		List<Point> GetBorderPointsFromFreeLine(Bitmap bitmap, List<Point> points)
		{
			Bitmap tempBitmap = new Bitmap(bitmap.Width, bitmap.Height);
			Graphics g = Graphics.FromImage(tempBitmap);
			g.Clear(Color.White);

			Point fromPoint = points[0];
			Point[] points1 = points.ToArray();

			foreach(Point toPoint in points1)
			{
				g.DrawLine(new Pen(Color.Black), fromPoint.X, fromPoint.Y, toPoint.X, toPoint.Y);
				fromPoint = toPoint;
			}

			g.Dispose();

			List<Point> borders = new List<Point>();

			for(int x = 0; x < tempBitmap.Width; x++)
			{
				for(int y = 0; y < tempBitmap.Height; y++)
				{
					Color color = tempBitmap.GetPixel(x, y);
					if(color.ToArgb() == Color.Black.ToArgb())
					{
						borders.Add(new Point(x, y));
					}
				}
			}
			tempBitmap.Dispose();
			return borders;
		}

		List<Point> borderPoints = new List<Point>();

		void OnMouseUpForMoveSelection(Point bitmapPoint)
		{
			isMouseDownForMove = false;
			if(BitmapRectangle != null)
				BitmapRectangle.UpdatePosition();
		}

		void OnMouseUpForSizeChangeSelection(Point bitmapPoint)
		{
			isMouseDownForSizeChange = false;
			BitmapRectangle.UpdatePosition();
		}


		BitmapRectangle blankBitmap = null;
		public void CreateBlankBitmap(Rectangle rect)
		{
			Bitmap bmp = Bitmap;

			if(blankBitmap != null)
				blankBitmap.Dispose();

			Bitmap bitmap = new Bitmap(rect.Width, rect.Height);
			for(int x = 0; x < rect.Width; x++)
			{
				if(x + rect.X > bmp.Width - 1)
					continue;

				for(int y = 0; y < rect.Height; y++)
				{
					if(y + rect.Y > bmp.Height - 1)
						continue;

					Color color = bmp.GetPixel(x + rect.X, y + rect.Y);
					if(color == Color.FromArgb(0, 0, 0, 0))
						bitmap.SetPixel(x, y, color);
					else
						bitmap.SetPixel(x, y, Color.White);
				}
			}

			blankBitmap = new BitmapRectangle(bitmap, rect);

			// BitmapRectangle.Y
			// rect.Y

			//MessageBox.Show(BitmapRectangle.Y.ToString() + " " + rect.Y.ToString());
		}

		void BitmapRectangle_MouseMoveForSizeChange(MouseEventArgs e)
		{
			int xMove = e.X + ScrollBarPosX - rectangleMoveStartMousePosX;
			int yMove = e.Y + ScrollBarPosY - rectangleMoveStartMousePosY;

			if(dragGriper == Griper.East)
				BitmapRectangle.Width = BitmapRectangle.OldWidth + xMove;
			if(dragGriper == Griper.South)
				BitmapRectangle.Height = BitmapRectangle.OldHeight + yMove;
			if(dragGriper == Griper.SouthEast)
			{
				BitmapRectangle.Width = BitmapRectangle.OldWidth + xMove;
				BitmapRectangle.Height = BitmapRectangle.OldHeight + yMove;
			}
			if(dragGriper == Griper.North)
			{
				BitmapRectangle.Y = BitmapRectangle.OldY + yMove;
				BitmapRectangle.Height = BitmapRectangle.OldHeight - yMove;
			}
			if(dragGriper == Griper.West)
			{
				BitmapRectangle.X = BitmapRectangle.OldX + xMove;
				BitmapRectangle.Width = BitmapRectangle.OldWidth - xMove;
			}
			if(dragGriper == Griper.NorthWest)
			{
				BitmapRectangle.X = BitmapRectangle.OldX + xMove;
				BitmapRectangle.Width = BitmapRectangle.OldWidth - xMove;
				BitmapRectangle.Y = BitmapRectangle.OldY + yMove;
				BitmapRectangle.Height = BitmapRectangle.OldHeight - yMove;
			}
			if(dragGriper == Griper.NorthEast)
			{
				BitmapRectangle.Width = BitmapRectangle.OldWidth + xMove;
				BitmapRectangle.Y = BitmapRectangle.OldY + yMove;
				BitmapRectangle.Height = BitmapRectangle.OldHeight - yMove;
			}
			if(dragGriper == Griper.SouthWest)
			{
				BitmapRectangle.X = BitmapRectangle.OldX + xMove;
				BitmapRectangle.Width = BitmapRectangle.OldWidth - xMove;
				BitmapRectangle.Height = BitmapRectangle.OldHeight + yMove;
			}
			pictureBox1.Invalidate();
		}

		void OnMouseMoveForSelectRectangle(Point bitmapPoint)
		{
			endPoint = bitmapPoint;
			ShowBitmap(Bitmap);
		}

		void OnMouseMoveForDrawRectangle(Point bitmapPoint)
		{
			endPoint = bitmapPoint;
			InvalidatePictureBox();
		}

		void OnMouseMoveForDrawLine(Point bitmapPoint)
		{
			endPoint = bitmapPoint;
			InvalidatePictureBox();
		}

		void OnMouseMoveForDrawFreeLine(Point bitmapPoint)
		{
			endPoint = bitmapPoint;

			linePoints.Add(bitmapPoint);
			InvalidatePictureBox();
		}


		void ChangeCursorWhenMouseMoveOnSelection(Point bitmapPoint)
		{
			if(BitmapRectangle != null)
			{
				Griper griper = IsMouseOnGriper(bitmapPoint);
				if(griper == Griper.Out)
					Cursor = Cursors.Default;
				if(griper == Griper.In)
					Cursor = Cursors.Hand;
				if(griper == Griper.NorthWest)
					Cursor = Cursors.SizeNWSE;
				if(griper == Griper.West)
					Cursor = Cursors.SizeWE;
				if(griper == Griper.SouthWest)
					Cursor = Cursors.SizeNESW;
				if(griper == Griper.NorthEast)
					Cursor = Cursors.SizeNESW;
				if(griper == Griper.East)
					Cursor = Cursors.SizeWE;
				if(griper == Griper.SouthEast)
					Cursor = Cursors.SizeNWSE;
				if(griper == Griper.North)
					Cursor = Cursors.SizeNS;
				if(griper == Griper.South)
					Cursor = Cursors.SizeNS;
			}
			else
			{
				int harfWidth = Bitmap.Width / 2;
				int harfHeight = Bitmap.Height / 2;

				if(harfWidth -4 <= bitmapPoint.X && bitmapPoint.X <= harfWidth + 4 &&
					Bitmap.Height -8 <= bitmapPoint.Y && bitmapPoint.Y <= Bitmap.Height )
					Cursor = Cursors.SizeNS;
				else if(harfHeight - 4 <= bitmapPoint.Y && bitmapPoint.Y <= harfHeight + 4 &&
					Bitmap.Width - 8 <= bitmapPoint.X && bitmapPoint.X <= Bitmap.Width )
					Cursor = Cursors.SizeWE;
				else if(Bitmap.Width - 8 <= bitmapPoint.X && bitmapPoint.X <= Bitmap.Width &&
					Bitmap.Height - 8 <= bitmapPoint.Y && bitmapPoint.Y <= Bitmap.Height )
					Cursor = Cursors.SizeNWSE;
				else
					Cursor = Cursors.Default;
			}
		}

		const int GriperSize = 8;
		Griper IsMouseOnGriper(Point bitmapPoint)
		{
			if(BitmapRectangle == null)
			{
				return Griper.Out;
			}

			Rectangle rect = BitmapRectangle.Rectangle;
			if(IsPointInRectangle(new Point(bitmapPoint.X, bitmapPoint.Y), rect))
			{
				if(rect.X <= bitmapPoint.X && bitmapPoint.X <= rect.X + GriperSize)
				{
					if(rect.Y <= bitmapPoint.Y && bitmapPoint.Y <= rect.Y + GriperSize)
						return Griper.NorthWest;
					if(rect.Bottom - GriperSize <= bitmapPoint.Y && bitmapPoint.Y <= rect.Bottom)
						return Griper.SouthWest;
					if(rect.Y + rect.Height/2 - GriperSize/2 <= bitmapPoint.Y && bitmapPoint.Y <= rect.Y + rect.Height / 2 + GriperSize/2)
						return Griper.West;
				}
				if(rect.Right - GriperSize <= bitmapPoint.X && bitmapPoint.X <= rect.Right)
				{
					if(rect.Y <= bitmapPoint.Y && bitmapPoint.Y <= rect.Y + GriperSize)
						return Griper.NorthEast;
					if(rect.Bottom - GriperSize <= bitmapPoint.Y && bitmapPoint.Y <= rect.Bottom)
						return Griper.SouthEast;
					if(rect.Y + rect.Height / 2 - GriperSize/2 <= bitmapPoint.Y && bitmapPoint.Y <= rect.Y + rect.Height / 2 + GriperSize/2)
						return Griper.East;
				}
				if(rect.X + rect.Width/2 - GriperSize/2 <= bitmapPoint.X && bitmapPoint.X <= rect.X + rect.Width / 2 + GriperSize/2)
				{
					if(rect.Y <= bitmapPoint.Y && bitmapPoint.Y <= rect.Y + GriperSize)
						return Griper.North;
					if(rect.Bottom - GriperSize <= bitmapPoint.Y && bitmapPoint.Y <= rect.Bottom)
						return Griper.South;
				}
				return Griper.In;
			}
			else
				return Griper.Out;
		}

		Griper IsMouseOnFirldGriper(Point bitmapPoint)
		{
			if(BitmapRectangle == null)
			{
				int harfWidth = Bitmap.Width / 2;
				int harfHeight = Bitmap.Height / 2;

				if(harfWidth - GriperSize / 2 <= bitmapPoint.X && bitmapPoint.X <= harfWidth + GriperSize / 2 &&
					Bitmap.Height - GriperSize <= bitmapPoint.Y && bitmapPoint.Y <= Bitmap.Height)
					return Griper.South;
				else if(harfHeight - GriperSize / 2 <= bitmapPoint.Y && bitmapPoint.Y <= harfHeight + GriperSize / 2 &&
					Bitmap.Width - GriperSize <= bitmapPoint.X && bitmapPoint.X <= Bitmap.Width)
					return Griper.East;
				else if(Bitmap.Width - GriperSize <= bitmapPoint.X && bitmapPoint.X <= Bitmap.Width &&
					Bitmap.Height - GriperSize <= bitmapPoint.Y && bitmapPoint.Y <= Bitmap.Height)
					return Griper.SouthEast;
				else
					return Griper.Out;
			}
			return Griper.Out;
		}

		public void ClearScrollInfo()
		{
			DragDropRectangle = new Rectangle();
			startPoint = new Point(-1, -1);
			endPoint = new Point(-1, -1);

			ScrollBarPosX = 0;
			ScrollBarPosY = 0;
		}

		bool[,] isCheckedFillColor;

		public void FillColor(Point bitmapPoint, Color newColor)
		{
			if(Bitmap == null)
				return;

			Color oldColor = Bitmap.GetPixel(bitmapPoint.X, bitmapPoint.Y);

			Bitmap newBitmap = new Bitmap(Bitmap);
			isCheckedFillColor = new bool[newBitmap.Width, newBitmap.Height];

			if(oldColor.ToArgb() == newColor.ToArgb())
				return;

			var list = FillLine(newBitmap, bitmapPoint.X, bitmapPoint.Y, oldColor, newColor);
			List<Point> nextPoints = new List<Point>();

			while(true)
			{
				foreach(var pt in list)
					nextPoints.AddRange(FillLine(newBitmap, pt.X, pt.Y, oldColor, newColor));
				list = nextPoints.ToList();
				nextPoints.Clear();

				if(list.Count == 0)
					break;
			}
			Bitmap = newBitmap;
			ShowBitmap(Bitmap);
		}

		List<Point> FillLine(Bitmap bitmap, int x, int y, Color oldColor, Color newColor)
		{
			isCheckedFillColor[x, y] = true;
			List<Point> nextPoints = new List<Point>();

			if(oldColor.ToArgb() == newColor.ToArgb())
				return nextPoints;

			Color color = bitmap.GetPixel(x, y);

			if(color == oldColor)
			{
				bitmap.SetPixel(x, y, newColor);
				if(y - 1 >= 0 && !isCheckedFillColor[x, y - 1] && bitmap.GetPixel(x, y - 1) == oldColor)
					nextPoints.Add(new Point(x, y - 1));
				if(y + 1 < bitmap.Height && !isCheckedFillColor[x, y + 1] && bitmap.GetPixel(x, y + 1) == oldColor)
					nextPoints.Add(new Point(x, y + 1));
			}

			int tempX = x;
			while(true)
			{
				tempX--;
				if(tempX < 0)
					break;

				Color color1 = bitmap.GetPixel(tempX, y);
				if(color1 == oldColor)
				{
					bitmap.SetPixel(tempX, y, newColor);
					if(y - 1 >= 0 && !isCheckedFillColor[tempX, y - 1] && bitmap.GetPixel(tempX, y - 1) == oldColor)
						nextPoints.Add(new Point(tempX, y - 1));
					if(y + 1 < bitmap.Height && !isCheckedFillColor[tempX, y + 1] && bitmap.GetPixel(tempX, y + 1) == oldColor)
						nextPoints.Add(new Point(tempX, y + 1));
				}
				else
					break;
			}

			tempX = x;
			while(true)
			{
				tempX++;
				if(tempX >= bitmap.Width)
					break;

				Color color1 = bitmap.GetPixel(tempX, y);
				if(color1 == oldColor)
				{
					bitmap.SetPixel(tempX, y, newColor);
					if(y - 1 >= 0 && !isCheckedFillColor[tempX, y - 1] && bitmap.GetPixel(tempX, y - 1) == oldColor)
						nextPoints.Add(new Point(tempX, y - 1));
					if(y + 1 < bitmap.Height && !isCheckedFillColor[tempX, y + 1] && bitmap.GetPixel(tempX, y + 1) == oldColor)
						nextPoints.Add(new Point(tempX, y + 1));
				}
				else
					break;
			}

			if(nextPoints.Count == 0)
				return nextPoints;

			if(nextPoints.Count == nextPoints.Max(x1 => x1.X) - nextPoints.Min(x1 => x1.X) + 1)
			{
				List<Point> points = new List<Point>();
				points.Add(nextPoints[0]);
				return points;
			}
			return nextPoints;
		}

		public EditMode EditMode
		{
			set;
			get;
		}
		= EditMode.Selection;

		private void UserControlImage_Load(object sender, EventArgs e)
		{

		}
	}

	public enum EditMode
	{
		Selection,
		SelectionFree,
		DrawRectangle,
		FillRectangle,
		DrawEllipse,
		FillEllipse,
		DrawLine,
		FreeLine,
		FillColor,
	}

	public enum Griper
	{
		Out,
		In,
		North,
		East,
		South,
		West,
		NorthEast,
		NorthWest,
		SouthEast,
		SouthWest,
	}

	public class UndoBuf
	{
		public UndoBuf(Bitmap bitmap, Rectangle rect)
		{
			Bitmap = bitmap;
			Rect = rect;
		}

		public Bitmap Bitmap
		{
			get;
			private set;
		}

		public Rectangle Rect
		{
			get;
			private set;
		}
	}

	public class UndoBufs
	{
		List<UndoBuf> undoBufs = new List<UndoBuf>();
		List<UndoBuf> redoBufs = new List<UndoBuf>();

		public void InsertUndoBuf(UndoBuf undoBuf)
		{
			undoBufs.Insert(0, undoBuf);
		}

		public void InsertRedoBuf(UndoBuf redoBuf)
		{
			redoBufs.Insert(0, redoBuf);
		}

		public void ClearRedoBufs()
		{
			redoBufs.Clear();
		}

		public UndoBuf RemoveUndoBuf()
		{
			if(undoBufs.Count > 0)
			{
				UndoBuf undoBuf = undoBufs[0];
				undoBufs.RemoveAt(0);
				return undoBuf;
			}
			else
				return null;
		}

		public UndoBuf RemoveRedoBuf()
		{
			if(redoBufs.Count > 0)
			{
				UndoBuf redoBuf = redoBufs[0];
				redoBufs.RemoveAt(0);
				return redoBuf;
			}
			else
				return null;
		}

		public bool CanUndo()
		{
			return undoBufs.Count > 0;
		}

		public bool CanRedo()
		{
			return redoBufs.Count > 0;
		}

		public void ClearUndoBufs()
		{
			undoBufs.Clear();
			redoBufs.Clear();
		}

		public int UndoBufsCount
		{
			get { return undoBufs.Count; }
		}
		public int RedoBufsCount
		{
			get { return redoBufs.Count; }
		}
	}
}
