﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Reflection;
using System.Runtime.Remoting.Channels;

namespace PaintApp
{
	public partial class UserControlImage : UserControl
	{
		void OnMouseUpForSelectFreeArea(Point bitmapPoint)
		{
			isMouseDown = false;

			endPoint = bitmapPoint;
			linePoints.Add(bitmapPoint);

			// 自由曲線でかこまれた境界線を保存する
			borderPoints = GetBorderPointsFromFreeLine(Bitmap, linePoints);

			//linePoints.Clear();
			startPoint = new Point(-1, -1);
			endPoint = new Point(-1, -1);

			InvalidatePictureBox();
		}

		private void PictureBox1_MouseDown(object sender, MouseEventArgs e)
		{
			if(Bitmap == null)
				return;

			int x = (int)(e.X / DisplayMagnification);
			int y = (int)(e.Y / DisplayMagnification);

			// 範囲外におけるクリック
			if(x + ScrollBarPosX >= Bitmap.Width || y + ScrollBarPosY >= Bitmap.Height)
			{
				OnMouseDownOutOfBitmap();
				Console.WriteLine("OnMouseDownOutOfBitmap()");
				return;
			}

			// クリックされた場所の色を通知する
			if(PictureBoxMouseDown != null)
			{
				Color color = Bitmap.GetPixel(x + ScrollBarPosX, y + ScrollBarPosY);
				PictureBoxMouseDown(this, new PictureBox1MouseDownArgs(x + ScrollBarPosX, y + ScrollBarPosY, color));
			}

			if(BitmapRectangle == null)
			{
				if(EditMode == EditMode.Selection)
					OnMouseDownForSelectRectangle(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.DrawRectangle || EditMode == EditMode.FillRectangle || EditMode == EditMode.DrawEllipse || EditMode == EditMode.FillEllipse)
					OnMouseDownForDrawRectangle(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.DrawLine)
					OnMouseDownForDrawLine(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.FreeLine)
					OnMouseDownForDrawFreeLine(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.SelectionFree)
					OnMouseDownForSelectFreeArea(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
			}
			else
			{
				if(IsPointInRectangle(new Point(x + ScrollBarPosX, y + ScrollBarPosY), BitmapRectangle.Rectangle))
					OnMouseDownInBitmapRectangle(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				else
					OnMouseDownOutOfSelection(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
			}
		}

		public Color GetColorFromPoint(int x, int y)
		{
			if(blankBitmap != null)
			{
				int minX = blankBitmap.Rectangle.Left;
				int minY = blankBitmap.Rectangle.Top;
				if(blankBitmap.Rectangle.Left <= x && x < blankBitmap.Rectangle.Right && blankBitmap.Rectangle.Top <= y && y < blankBitmap.Rectangle.Bottom)
					return blankBitmap.Bitmap.GetPixel(x - minX, y - minY);
			}

			if(x < Bitmap.Width && y < Bitmap.Height)
				return Bitmap.GetPixel(x, y);
			else
				return Color.Empty;
		}

		public class PictureBoxMouseMoveEventArgs : EventArgs
		{
			public PictureBoxMouseMoveEventArgs(int x, int y)
			{
				X = x;
				Y = y;
			}

			public int X
			{
				get;
				protected set;
			}

			public int Y
			{
				get;
				protected set;
			}
		}
		public delegate void PictureBoxMouseMoveHandler(object sender, PictureBoxMouseMoveEventArgs e);
		public event PictureBoxMouseMoveHandler PictureBoxMouseMove;

		private void PictureBox1_MouseMove(object sender, MouseEventArgs e)
		{
			if(Bitmap == null)
				return;

			int x = (int)(e.X / DisplayMagnification);
			int y = (int)(e.Y / DisplayMagnification);

			PictureBoxMouseMove?.Invoke(this, new PictureBoxMouseMoveEventArgs(x + ScrollBarPosX, y + ScrollBarPosY));

			if(isMouseDown)
			{
				if(EditMode == EditMode.Selection)
					OnMouseMoveForSelectRectangle(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.DrawRectangle || EditMode == EditMode.FillRectangle || EditMode == EditMode.DrawEllipse || EditMode == EditMode.FillEllipse)
					OnMouseMoveForDrawRectangle(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.DrawLine)
					OnMouseMoveForDrawLine(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.FreeLine)
					OnMouseMoveForDrawFreeLine(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.SelectionFree)
					OnMouseMoveForSelectFreeArea(new Point(x + ScrollBarPosX, y + ScrollBarPosY));//tuika
			}
			else
				ChangeCursorWhenMouseMoveOnSelection(new Point(x + ScrollBarPosX, y + ScrollBarPosY));

			if(isMouseDownForMove)
			{
				if(EditMode == EditMode.SelectionFree)
					OnMouseMoveForMoveFreeLineSelection(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				else if(EditMode == EditMode.Selection)
					OnMouseMoveForMoveSelection(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				else if(EditMode == EditMode.DrawRectangle || EditMode == EditMode.FillRectangle || EditMode == EditMode.DrawEllipse || EditMode == EditMode.FillEllipse)
					OnMouseMoveForMoveSelection(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
			}

			if(isMouseDownForSizeChange)
				OnMouseMoveForSizeChangeSelection(new Point(x + ScrollBarPosX, y + ScrollBarPosY));

			if(isMouseDownForFieldSizeChange)
				OnMouseMoveForSizeChangeField(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
		}

		void DrawFieldGriper(Point bitmapPoint)
		{
			if(griperField == Griper.South || griperField == Griper.East || griperField == Griper.SouthEast)
			{
				int oldWidth = Bitmap.Width;
				int oldHeight = Bitmap.Height;

				int newWidth = bitmapPoint.X;
				int newHeight = bitmapPoint.Y;

				if(griperField == Griper.South)
					newWidth = oldWidth;
				if(griperField == Griper.East)
					newHeight = oldHeight;

				Bitmap bitmap = new Bitmap(newWidth, newHeight);
				Graphics g = Graphics.FromImage(bitmap);

				g.DrawImage(Bitmap, new Point(0, 0));
				g.DrawRectangle(new Pen(GetBoderBrush()), new Rectangle(-1, -1, bitmap.Width, bitmap.Height));

				int x = bitmap.Width - GriperSize;
				int y = bitmap.Height - GriperSize;
				g.DrawRectangle(new Pen(Color.Black), new Rectangle(x, y, GriperSize, GriperSize));

				x = bitmap.Width / 2 - GriperSize / 2;
				y = bitmap.Height - GriperSize;
				g.DrawRectangle(new Pen(Color.Black), new Rectangle(x, y, GriperSize, GriperSize));

				x = bitmap.Width - GriperSize;
				y = bitmap.Height / 2 - GriperSize / 2;
				g.DrawRectangle(new Pen(Color.Pink), new Rectangle(x, y, GriperSize, GriperSize));

				//g.DrawRectangle(new Pen(GetBoderBrush(), 10), new Rectangle(-1, -1, bitmap.Width + 1, bitmap.Height + 1));
				g.DrawLine(new Pen(GetBoderBrush(), 10), bitmap.Width + 1, 0, bitmap.Width + 1, bitmap.Height + 1);
				g.DrawLine(new Pen(GetBoderBrush(), 10), 0, bitmap.Height + 1, bitmap.Width + 1, bitmap.Height + 1);

				g.Dispose();

				OptimizationScrloolBar(bitmap);
				Bitmap bitmap1 = GetBitmapDrawPictureBox(bitmap);
				pictureBox1.Image = bitmap1;
			}
		}

		void OnMouseMoveForSizeChangeField(Point bitmapPoint)
		{
			DrawFieldGriper(bitmapPoint);
		}

		private void PictureBox1_MouseUp(object sender, MouseEventArgs e)
		{
			if(Bitmap == null)
				return;

			int x = (int)(e.X / DisplayMagnification);
			int y = (int)(e.Y / DisplayMagnification);

			if(isMouseDown)
			{
				if(EditMode == EditMode.Selection)
					OnMouseUpForSelectRectangle();
				if(EditMode == EditMode.DrawRectangle || EditMode == EditMode.FillRectangle || EditMode == EditMode.DrawEllipse || EditMode == EditMode.FillEllipse)
					OnMouseUpForDrawRectangle();
				if(EditMode == EditMode.DrawLine)
					OnMouseUpForDrawLine(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.FreeLine)
					OnMouseUpForDrawFreeLine(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				if(EditMode == EditMode.SelectionFree)
					OnMouseUpForSelectFreeArea(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				return;
			}

			if(isMouseDownForMove)
			{
				if(EditMode == EditMode.SelectionFree)
					OnMouseUpForMoveFreeLineSelection(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
				else
					OnMouseUpForMoveSelection(new Point(x + ScrollBarPosX, y + ScrollBarPosY));
			}
			if(isMouseDownForSizeChange)
				OnMouseUpForSizeChangeSelection(new Point(x + ScrollBarPosX, y + ScrollBarPosY));

			if(isMouseDownForFieldSizeChange)
				OnMouseUpForSizeChangeField(new Point(x + ScrollBarPosX, y + ScrollBarPosY));

			return;
		}

		void OnMouseUpForSizeChangeField(Point bitmapPoint)
		{
			Griper griper = griperField;

			isMouseDownForFieldSizeChange = false;
			startPoint = new Point();
			griperField = Griper.Out;

			if(griper == Griper.South || griper == Griper.East || griper == Griper.SouthEast)
			{
				int oldWidth = Bitmap.Width;
				int oldHeight = Bitmap.Height;
				int moveX = bitmapPoint.X - mousePositionStartFieldSizeChangeX;
				int moveY = bitmapPoint.Y - mousePositionStartFieldSizeChangeY;

				int newWidth = oldWidth + moveX;
				int newHeight = oldHeight + moveY;

				if(griper == Griper.South)
					newWidth = oldWidth;
				if(griper == Griper.East)
					newHeight = oldHeight;

				Bitmap newBitmap = new Bitmap(newWidth, newHeight);
				Graphics g = Graphics.FromImage(newBitmap);
				g.DrawImage(Bitmap, new Point(0, 0));
				g.Dispose();

				Bitmap = newBitmap;
			}
			ShowBitmap(Bitmap);
		}
	}

	public class RichTextBoxEx : RichTextBox
	{
		protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
		{
			if(keyData == Keys.A)
			{
				// ここにやりたいことを書く
				//MessageBox.Show("");
				return true;
			}

			return base.ProcessCmdKey(ref msg, keyData);
		}

	}
}

