﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Reflection;

namespace PaintApp
{
	public enum PasteMode1
	{
		Default,
		LeftSide,
		RightSide,
		UpperSide,
		UnderSide,
	}

	public partial class UserControlImage : UserControl
    {
		private void PictureBox1_Paint(object sender, PaintEventArgs e)
		{
			if(isMouseDown || BitmapRectangle != null)
			{
				ShowBitmap(e.Graphics, Bitmap);

				if(EditMode == EditMode.Selection)
				{
					DrawBoderRectangle(e.Graphics);
				}
				else if(EditMode == EditMode.SelectionFree)
				{
					if(isMouseDown)
						DrawFreeLine(e.Graphics, linePoints, new Pen(GetBoderBrush()));
					else
						DrawBoderRectangle(e.Graphics);
				}
				else if(EditMode == EditMode.DrawRectangle || EditMode == EditMode.FillRectangle || EditMode == EditMode.DrawEllipse || EditMode == EditMode.FillEllipse)
					DrawRectangle(e.Graphics);
				else if(EditMode == EditMode.FreeLine)
					DrawFreeLine(e.Graphics, linePoints, new Pen(GetBoderBrush()));
				else if(EditMode == EditMode.DrawLine)
					DrawLine(e.Graphics);
				else
					DrawBoderRectangle(e.Graphics);

				e.Graphics.DrawRectangle(new Pen(Color.Black), new Rectangle(-1, -1, Bitmap.Width - ScrollBarPosX + 1, Bitmap.Height - ScrollBarPosY + 1));
			}
			else
			{
				if(Bitmap != null)
				{
					if(isMouseDownForFieldSizeChange)
						return;

					ShowBitmap(e.Graphics, Bitmap);

					if(EditMode == EditMode.SelectionFree)
					{
						if(linePoints != null)
							DrawFreeLine(e.Graphics, linePoints, new Pen(GetBoderBrush()));
					}

					DrawFieldGrip(Bitmap, e.Graphics);
				}
			}
			return;
		}

		// OnMouseMoveForSelectRectangleによって呼ばれる
		void DrawBoderRectangle(Graphics g)
		{
			Rectangle rect;
			if(BitmapRectangle == null)
				rect = GetRectangle(startPoint, endPoint);
			else
				rect = BitmapRectangle.Rectangle;

			rect.X -= ScrollBarPosX;
			rect.Y -= ScrollBarPosY;

			rect.X = (int)(rect.X * DisplayMagnification);
			rect.Y = (int)(rect.Y * DisplayMagnification);
			rect.Width = (int)(rect.Width * DisplayMagnification);
			rect.Height = (int)(rect.Height * DisplayMagnification);

			g.DrawRectangle(new Pen(GetBoderBrush()), rect);

			List<Rectangle> griperRects = new List<Rectangle>();
			griperRects.Add(new Rectangle(new Point(rect.X, rect.Y), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.X + rect.Width / 2 - GriperSize / 2, rect.Y), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.Right - GriperSize, rect.Y), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.X, rect.Bottom - GriperSize), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.X + rect.Width / 2 - GriperSize / 2, rect.Bottom - GriperSize), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.Right - GriperSize, rect.Bottom - GriperSize), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.X, rect.Y + rect.Height / 2 - GriperSize / 2), new Size(GriperSize, GriperSize)));
			griperRects.Add(new Rectangle(new Point(rect.Right - GriperSize, rect.Y + rect.Height / 2 - GriperSize / 2), new Size(GriperSize, GriperSize)));
			g.FillRectangles(GetBoderBrush(), griperRects.ToArray());
		}

		void DrawLine(Graphics g)
		{
			int x1 = startPoint.X - ScrollBarPosX;
			int y1 = startPoint.Y - ScrollBarPosY;
			int x2 = endPoint.X - ScrollBarPosX;
			int y2 = endPoint.Y - ScrollBarPosY;

			x1 = (int)(x1 * DisplayMagnification);
			x2 = (int)(x2 * DisplayMagnification);
			y1 = (int)(y1 * DisplayMagnification);
			y2 = (int)(y2 * DisplayMagnification);

			g.DrawLine(new Pen(DrawColor), x1, y1, x2, y2);
		}

		void DrawFreeLine(Graphics g, List<Point> points, Pen pen)
		{
			if(points.Count < 2)
				return;

			Point fromPoint = points[0];
			Point[] points1 = points.ToArray();

			int xScrollBarPos = ScrollBarPosX;
			int yScrollBarPos = ScrollBarPosY;

			foreach(Point toPoint in points1)
			{
				int x1 = fromPoint.X - xScrollBarPos;
				int y1 = fromPoint.Y - yScrollBarPos;
				int x2 = toPoint.X - xScrollBarPos;
				int y2 = toPoint.Y - yScrollBarPos;

				x1 = (int)(x1 * DisplayMagnification);
				x2 = (int)(x2 * DisplayMagnification);
				y1 = (int)(y1 * DisplayMagnification);
				y2 = (int)(y2 * DisplayMagnification);

				g.DrawLine(pen, x1, y1, x2, y2);
				fromPoint = toPoint;
			}
		}

		void DrawRectangle(Graphics g)
		{
			Rectangle rect;
			if(BitmapRectangle == null)
				rect = GetRectangle(startPoint, endPoint);
			else
				rect = BitmapRectangle.Rectangle;

			rect.X -= ScrollBarPosX;
			rect.Y -= ScrollBarPosY;

			rect.X = (int)(rect.X * DisplayMagnification);
			rect.Y = (int)(rect.Y * DisplayMagnification);
			rect.Width = (int)(rect.Width * DisplayMagnification);
			rect.Height = (int)(rect.Height * DisplayMagnification);

			if(EditMode == EditMode.DrawRectangle)
				g.DrawRectangle(new Pen(DrawColor), rect);
			if(EditMode == EditMode.FillRectangle)
				g.FillRectangle(new SolidBrush(DrawColor), rect);
			if(EditMode == EditMode.DrawEllipse)
				g.DrawEllipse(new Pen(DrawColor), rect);
			if(EditMode == EditMode.FillEllipse)
				g.FillEllipse(new SolidBrush(DrawColor), rect);

			DrawBoderRectangle(g);
		}

		public void ShowBitmap(Graphics g, Bitmap bitmap)
		{
			OptimizationScrloolBar(bitmap);

			Bitmap bitmap1 = GetBitmapDrawPictureBox(bitmap);

			if(BitmapRectangle != null)
			{
				g.DrawImage(bitmap1, new Rectangle(0, 0, pictureBox1.Width, pictureBox1.Height));

				if(blankBitmap != null)
				{
					Rectangle rect = blankBitmap.Rectangle;

					rect.X -= ScrollBarPosX;
					rect.Y -= ScrollBarPosY;

					rect.X = (int)(rect.X * DisplayMagnification);
					rect.Y = (int)(rect.Y * DisplayMagnification);
					rect.Width = (int)(rect.Width * DisplayMagnification);
					rect.Height = (int)(rect.Height * DisplayMagnification);

					g.DrawImage(blankBitmap.Bitmap, rect);
				}

				if(blankColorPoints != null)
				{
					int x1 = blankColorPoints.Min(x => x.X);
					int y1 = blankColorPoints.Min(x => x.Y);
					int x2 = blankColorPoints.Max(x => x.X);
					int y2 = blankColorPoints.Max(x => x.Y);

					Bitmap blank = new Bitmap(x2 - x1 + 1, y2 - y1 + 1);
					foreach(var cp in blankColorPoints)
						blank.SetPixel(cp.X - x1, cp.Y - y1, cp.Color);

					Rectangle rect = new Rectangle(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
					blankBitmap = new BitmapRectangle(blank, rect);

					blankColorPoints.Clear();
					blankColorPoints = null;
				}

				{
					Rectangle rect = BitmapRectangle.Rectangle;

					rect.X -= ScrollBarPosX;
					rect.Y -= ScrollBarPosY;

					rect.X = (int)(rect.X * DisplayMagnification);
					rect.Y = (int)(rect.Y * DisplayMagnification);
					rect.Width = (int)(rect.Width * DisplayMagnification);
					rect.Height = (int)(rect.Height * DisplayMagnification);

					g.DrawImage(BitmapRectangle.Bitmap, rect);
				}
			}
			else
			{
				g.DrawImage(bitmap1, new Rectangle(0, 0, pictureBox1.Width, pictureBox1.Height));
			}
		}

		public void InvalidatePictureBox()
		{
			if(Bitmap == null)
				return;

			int bitmapWidth = (int)(Bitmap.Width * DisplayMagnification);
			int bitmapHeight = (int)(Bitmap.Height * DisplayMagnification);

			if(pictureBox1.Width >= bitmapWidth || pictureBox1.Height >= bitmapHeight)
				pictureBox1.Invalidate(new Rectangle(0, 0, bitmapWidth, bitmapHeight));
			else
				pictureBox1.Invalidate();
		}

		public PasteMode1 PasteMode1
		{
			get;
			set;
		}
		= PasteMode1.Default;

		public void PasteImage()
		{
			if(!Clipboard.ContainsImage())
				return;

			if(BitmapRectangle != null)
				UniteBitmapRectangle(new Point(0, 0));

			Bitmap image = (Bitmap)Clipboard.GetDataObject().GetData("System.Drawing.Bitmap");

			if(IsAutoSizePaste)
				image = GetBitmapAdjustedSize(image);

			PasteImage0(image);
		}

		Bitmap GetBitmapAdjustedSize(Bitmap image)
		{
			int oldWidth = Bitmap.Width;
			int oldHeight = Bitmap.Height;

			int newWidth = image.Width;
			int newHeight = image.Height;

			if(PasteMode1 == PasteMode1.UpperSide || PasteMode1 == PasteMode1.UnderSide)
			{
				newWidth = newWidth * oldWidth / image.Width;
				newHeight = newHeight * oldWidth / image.Width;
			}

			if(PasteMode1 == PasteMode1.LeftSide || PasteMode1 == PasteMode1.RightSide)
			{
				newWidth = newWidth * oldHeight / image.Height;
				newHeight = newHeight * oldHeight / image.Height;
			}

			Bitmap newBitmap = new Bitmap(newWidth, newHeight);
			Graphics g = Graphics.FromImage(newBitmap);
			Rectangle srcRect = new Rectangle(0, 0, image.Width, image.Height);
			Rectangle destRect = new Rectangle(0, 0, newWidth, newHeight);
			g.DrawImage(image, destRect, srcRect, GraphicsUnit.Pixel);
			g.Dispose();
			return newBitmap;
		}

		public void PasteImage0(Bitmap image)
		{
			int oldWidth = Bitmap.Width;
			int oldHeight = Bitmap.Height;

			if(PasteMode1 == PasteMode1.Default)
				BitmapRectangle = new BitmapRectangle(image, new Rectangle(new Point(ScrollBarPosX, ScrollBarPosY), image.Size));
			if(PasteMode1 == PasteMode1.UnderSide)
				BitmapRectangle = new BitmapRectangle(image, new Rectangle(new Point(0, Bitmap.Height), image.Size));
			if(PasteMode1 == PasteMode1.RightSide)
				BitmapRectangle = new BitmapRectangle(image, new Rectangle(new Point(Bitmap.Width, 0), image.Size));
			if(PasteMode1 == PasteMode1.LeftSide || PasteMode1 == PasteMode1.UpperSide)
				BitmapRectangle = new BitmapRectangle(image, new Rectangle(new Point(0, 0), image.Size));

			int newWidth = 0;
			int newHeight = 0;

			if(PasteMode1 == PasteMode1.LeftSide || PasteMode1 == PasteMode1.RightSide)
			{
				newWidth = oldWidth + image.Size.Width;
				newHeight = oldHeight;
				if(oldHeight < image.Size.Height)
					newHeight = image.Size.Height;
			}
			if(PasteMode1 == PasteMode1.UpperSide || PasteMode1 == PasteMode1.UnderSide)
			{
				newHeight = oldHeight + image.Size.Height;
				newWidth = oldWidth;
				if(oldWidth < image.Size.Width)
					newWidth = image.Size.Width;
			}
			if(PasteMode1 == PasteMode1.Default)
			{
				newWidth = oldWidth;
				newHeight = oldHeight;
				if(oldWidth < image.Size.Width)
					newWidth = image.Size.Width;
				if(oldHeight < image.Size.Height)
					newHeight = image.Size.Height;
			}

			Bitmap newBitmap = new Bitmap(newWidth, newHeight);

			Graphics g = Graphics.FromImage(newBitmap);
			if(PasteMode1 == PasteMode1.RightSide || PasteMode1 == PasteMode1.UnderSide || PasteMode1 == PasteMode1.Default)
				g.DrawImage(Bitmap, new Point(0, 0));
			if(PasteMode1 == PasteMode1.LeftSide)
				g.DrawImage(Bitmap, new Point(BitmapRectangle.Width, 0));
			if(PasteMode1 == PasteMode1.UpperSide)
				g.DrawImage(Bitmap, new Point(0, BitmapRectangle.Height));

			g.Dispose();

			Bitmap = newBitmap;
			this.InvalidatePictureBox();
		}

	}
}
